# Copyright (c) 2021 PaddlePaddle Authors. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
import os
import shutil

import numpy as np
import paddle
import paddle.nn as nn
import paddle.nn.functional as F

from paddlenlp.ops import (
    InferBartDecoding,
    InferGptDecoding,
    InferGptJDecoding,
    InferMBartDecoding,
    InferMIRODecoding,
    InferOptDecoding,
    InferPegasusDecoding,
    InferT5Decoding,
    InferTransformerDecoding,
    InferUnifiedDecoding,
)
from paddlenlp.transformers import (
    BartPretrainedModel,
    CodeGenPreTrainedModel,
    GPTChineseTokenizer,
    GPTJPretrainedModel,
    GPTPretrainedModel,
    GPTTokenizer,
    InferTransformerModel,
    MBartPretrainedModel,
    OPTPretrainedModel,
    PegasusPretrainedModel,
    PositionalEmbedding,
    T5PretrainedModel,
    TransformerModel,
    UnifiedTransformerPretrainedModel,
    UNIMOPretrainedModel,
    WordEmbedding,
    position_encoding_init,
)
from paddlenlp.utils.log import logger

from .encoder import enable_fast_encoder


class FasterTransformer(TransformerModel):
    """
    FasterTransformer is a fast version for generation with the Transformer
    model. It uses a custom op based on and enhancing NV FasterTransformer to
    do fast generation.

    Args:
        src_vocab_size (int):
            The size of source vocabulary.
        trg_vocab_size (int):
            The size of target vocabulary.
        max_length (int):
            The maximum length of input sequences.
        num_encoder_layers (int):
            The number of sub-layers to be stacked in the encoder.
        num_decoder_layers (int):
            The number of sub-layers to be stacked in the decoder.
        n_head (int):
            The number of head used in multi-head attention.
        d_model (int):
            The dimension for word embeddings, which is also the last dimension of
            the input and output of multi-head attention, position-wise feed-forward
            networks, encoder and decoder.
        d_inner_hid (int):
            Size of the hidden layer in position-wise feed-forward networks.
        dropout (float):
            Dropout rates. Used for pre-process, activation and inside attention.
        weight_sharing (bool):
            Whether to use weight sharing.
        attn_dropout (float):
            The dropout probability used in MHA to drop some attention target.
            If None, use the value of dropout. Defaults to None.
        act_dropout (float):
            The dropout probability used after FFN activition. If None, use
            the value of dropout. Defaults to None.
        bos_id (int, optional):
            The start token id and also is used as padding id. Defaults to 0.
        eos_id (int, optional):
            The end token id. Defaults to 1.
        pad_id (int, optional):
            The pad token id. Defaults to None. If it's None, the bos_id will be used as pad_id.
        decoding_strategy (str, optional):
            Indicating the strategy of decoding. It can be 'beam_search', 'beam_search_v2',
            'topk_sampling' and 'topp_sampling'. For beam search strategies,
            'v2' would select the top `beam_size * 2` beams and process the top
            `beam_size` alive and finish beams in them separately, while 'v1'
            would only select the top `beam_size` beams and mix up the alive and
            finish beams. 'v2' always searchs more and get better results, since
            the alive beams would always be `beam_size` while the number of alive
            beams in `v1` might decrease when meeting the end token. However,
            'v2' always generates longer results thus might do more calculation
            and be slower.
        beam_size (int, optional):
            The beam width for beam search. Defaults to 4.
        topk (int, optional):
            The number of highest probability tokens to keep for top-k sampling.
            Defaults to 4.
        topp (float, optional):
            The most probable tokens whose cumulative probability is not less than
            `topp` are kept for top-p sampling. Defaults to 4.
        max_out_len (int, optional):
            The maximum output length. Defaults to 256.
        diversity_rate (float, optional):
            Refer to `A Simple, Fast Diverse Decoding Algorithm for Neural Generation <https://arxiv.org/abs/1611.08562>`_
            for details. Bigger `diversity_rate` would lead to more diversity.
            if `diversity_rate == 0` is equivalent to naive BeamSearch. Default
            to 0 if not set.
        use_fp16_decoding(bool, optional):
            Whether to use fp16 for decoding.
        enable_fast_encoder(bool, optional):
            Whether to use the fast version of encoder. This is experimental option for now.
            Defaults to False.
        use_fp16_encoder(bool, optional):
            Whether to use fp16 for encoder. Only works when enable_fast_encoder is True.
            Defaults to False.
        rel_len(bool, optional):
            Indicating whether `max_out_len` in is the length relative to that
            of source text. Only works in `v2` temporarily. It is suggest to set
            a small `max_out_len` and use `rel_len=True`. Default to False if
            not set.
        alpha(float, optional):
            The power number in length penalty calculation. Only works in `v2`
            temporarily. Refer to `GNMT <https://arxiv.org/pdf/1609.08144.pdf>`_.
            Default to 0.6 if not set.
    """

    def __init__(
        self,
        src_vocab_size,
        trg_vocab_size,
        max_length,
        num_encoder_layers,
        num_decoder_layers,
        n_head,
        d_model,
        d_inner_hid,
        dropout,
        weight_sharing,
        attn_dropout=None,
        act_dropout=None,
        bos_id=0,
        eos_id=1,
        pad_id=None,
        decoding_strategy="beam_search",
        beam_size=4,
        topk=1,
        topp=0.0,
        max_out_len=256,
        diversity_rate=0.0,
        decoding_lib=None,
        use_fp16_decoding=False,
        enable_fast_encoder=False,
        use_fp16_encoder=False,
        rel_len=False,
        alpha=0.6,
    ):
        # if decoding_lib is None:
        #     raise ValueError(
        #         "The args decoding_lib must be set to use FasterTransformer. ")
        # elif not os.path.exists(decoding_lib):
        #     raise ValueError("The path to decoding lib is not exist.")

        args = dict(locals())
        args.pop("self")
        args.pop("__class__", None)
        self.decoding_strategy = args.pop("decoding_strategy")
        self.beam_size = args.pop("beam_size")
        self.topk = args.pop("topk")
        self.topp = args.pop("topp")
        self.max_out_len = args.pop("max_out_len")
        self.diversity_rate = args.pop("diversity_rate")
        self.decoding_lib = args.pop("decoding_lib")
        self.use_fp16_decoding = args.pop("use_fp16_decoding")
        self.enable_fast_encoder = args.pop("enable_fast_encoder")
        self.use_fp16_encoder = args.pop("use_fp16_encoder")
        self.rel_len = args.pop("rel_len")
        self.alpha = args.pop("alpha")
        self.dropout = dropout
        self.weight_sharing = weight_sharing
        self.trg_vocab_size = trg_vocab_size
        self.d_model = d_model
        self.bos_id = bos_id
        self.pad_id = pad_id if pad_id is not None else self.bos_id
        self.max_length = max_length
        super(FasterTransformer, self).__init__(**args)

        if self.enable_fast_encoder:
            logger.warning("enable_fast_encoder is an experimental option and subject to change.")
        elif self.use_fp16_encoder:
            self.use_fp16_encoder = False

        self.decoding_linear = nn.Linear(in_features=d_model, out_features=trg_vocab_size)

        if weight_sharing:
            self.trg_word_embedding = WordEmbedding(vocab_size=trg_vocab_size, emb_dim=d_model, bos_id=self.bos_id)
            self.trg_pos_embedding = PositionalEmbedding(emb_dim=d_model, max_length=max_length)

        self.decoding = InferTransformerDecoding(
            decoder=self.transformer.decoder,
            word_embedding=self.trg_word_embedding.word_embedding,
            positional_embedding=self.trg_pos_embedding.pos_encoder,
            linear=self.decoding_linear,
            num_decoder_layers=num_decoder_layers,
            n_head=n_head,
            d_model=d_model,
            bos_id=bos_id,
            eos_id=eos_id,
            decoding_strategy=decoding_strategy,
            beam_size=beam_size,
            topk=topk,
            topp=topp,
            max_out_len=max_out_len,
            diversity_rate=self.diversity_rate,
            decoding_lib=self.decoding_lib,
            use_fp16_decoding=self.use_fp16_decoding,
            rel_len=self.rel_len,
            alpha=self.alpha,
        )

    def forward(self, src_word, trg_word=None):
        src_max_len = paddle.shape(src_word)[-1]
        src_slf_attn_bias = (
            paddle.cast(src_word == self.pad_id, dtype=paddle.get_default_dtype()).unsqueeze([1, 2]) * -1e9
        )
        src_pos = paddle.cast(src_word != self.pad_id, dtype=src_word.dtype) * paddle.arange(start=0, end=src_max_len)

        # Run encoder
        src_emb = self.src_word_embedding(src_word)
        src_pos_emb = self.src_pos_embedding(src_pos)
        src_emb = src_emb + src_pos_emb
        enc_input = F.dropout(src_emb, p=self.dropout, training=False) if self.dropout else src_emb

        if self.enable_fast_encoder and self.use_fp16_encoder:
            enc_input = paddle.cast(enc_input, dtype="float16")

        enc_output = self.transformer.encoder(enc_input, src_slf_attn_bias)

        if self.use_fp16_decoding and enc_output.dtype != paddle.float16:
            enc_output = paddle.cast(enc_output, dtype="float16")
        elif not self.use_fp16_decoding and enc_output.dtype != paddle.float32:
            enc_output = paddle.cast(enc_output, dtype="float32")

        mem_seq_lens = paddle.sum(paddle.cast(src_word != self.pad_id, dtype="int32"), dtype="int32", axis=1)
        ids = self.decoding(enc_output, mem_seq_lens, trg_word=trg_word)

        return ids

    def load(self, init_from_params=None, state_dict=None):
        # Load the trained model
        if init_from_params is None and state_dict is None:
            raise ValueError("Either init_from_params or state_dict must be given to load the infer model. ")

        if state_dict is None:
            state_dict = paddle.load(init_from_params, return_numpy=True)
        else:
            for state in state_dict:
                # NOTE: This API only used in dygraph, so paddle.Tensor is enough.
                if isinstance(state_dict[state], paddle.Tensor):
                    state_dict[state] = state_dict[state].numpy()

        # To set weight[padding_idx] to 0.
        state_dict["trg_word_embedding.word_embedding.weight"][self.bos_id] = [0] * self.d_model

        # Dealing with weight sharing.
        if self.weight_sharing:
            state_dict["decoding_linear.weight"] = np.transpose(state_dict["trg_word_embedding.word_embedding.weight"])
        else:
            state_dict["decoding_linear.weight"] = state_dict["linear.weight"]

        if self.decoding._fuse_qkv:
            for item in self.state_dict():
                if "decoder" in item and "self_attn.q_proj" in item:
                    num_layer = item.split(".")[3]
                    param_type = item.split(".")[-1]

                    state_dict["decoding.slf_q_" + param_type + "_" + num_layer] = np.concatenate(
                        (
                            state_dict[item],
                            state_dict["transformer.decoder.layers." + num_layer + ".self_attn.k_proj." + param_type],
                            state_dict["transformer.decoder.layers." + num_layer + ".self_attn.v_proj." + param_type],
                        ),
                        axis=-1,
                    )

        if self.use_fp16_decoding:
            for item in self.state_dict():
                if "decoder" in item or "decoding.slf" in item:
                    state_dict[item] = np.float16(state_dict[item])
            state_dict["decoding_linear.weight"] = np.float16(state_dict["decoding_linear.weight"])
            state_dict["trg_word_embedding.word_embedding.weight"] = np.float16(
                state_dict["trg_word_embedding.word_embedding.weight"]
            )
            state_dict["trg_pos_embedding.pos_encoder.weight"] = np.float16(
                state_dict["trg_pos_embedding.pos_encoder.weight"]
            )
            state_dict["decoding_linear.bias"] = np.zeros([self.trg_vocab_size], dtype="float16")

        self.load_dict(state_dict)

        if self.enable_fast_encoder:
            self = enable_fast_encoder(self, use_fp16=self.use_fp16_encoder)

    def export_params(self, init_from_params, place):
        """
        This method is used for load static graph from dygraph checkpoint
        or export inference model using static graph.
        Do NOT support faster encoder.

        Args:
            init_from_params (string):
                The path to dygraph checkpoint.
            place (paddle.Place):
                The place to execute static graph.

        Example:
            .. code-block::
                paddle.enable_static()
                place = "gpu"
                place = paddle.set_device(place)
                reader.adapt_vocab_size(args)

                test_program = paddle.static.Program()
                startup_program = paddle.static.Program()
                with paddle.static.program_guard(test_program, startup_program):
                    src_word = paddle.static.data(
                        name="src_word", shape=[None, None], dtype="int64")

                    # Define model
                    transformer = FasterTransformer(
                        src_vocab_size=args.src_vocab_size,
                        trg_vocab_size=args.trg_vocab_size,
                        max_length=args.max_length + 1,
                        num_encoder_layers=args.n_layer,
                        num_decoder_layers=args.n_layer,
                        n_head=args.n_head,
                        d_model=args.d_model,
                        d_inner_hid=args.d_inner_hid,
                        dropout=args.dropout,
                        weight_sharing=args.weight_sharing,
                        bos_id=args.bos_idx,
                        eos_id=args.eos_idx,
                        decoding_strategy=args.decoding_strategy,
                        beam_size=args.beam_size,
                        max_out_len=args.max_out_len,
                        decoding_lib=args.decoding_lib,
                        use_fp16_decoding=args.use_fp16_decoding,
                        rel_len=args.use_rel_len,
                        alpha=args.alpha)

                    finished_seq = transformer(src_word=src_word)

                test_program = test_program.clone(for_test=True)

                exe = paddle.static.Executor(place)
                exe.run(startup_program)

                # Load checkpoint.
                transformer.export_params(
                    init_from_params=os.path.join(args.init_from_params,
                                                "transformer.pdparams"),
                    place=place)

                paddle.static.save_inference_model(
                    os.path.join(args.inference_model_dir, "transformer"),
                    feed_vars=src_word,
                    fetch_vars=finished_seq,
                    executor=exe,
                    program=test_program)
        """
        # Load the trained model
        assert init_from_params, "Please set init_from_params to load the infer model."

        model_dict = paddle.load(init_from_params, return_numpy=True)

        # To set weight[padding_idx] to 0.
        model_dict["trg_word_embedding.word_embedding.weight"][self.bos_id] = [0] * self.d_model

        # Dealing with weight sharing.
        if self.weight_sharing:
            model_dict["decoding_linear.weight"] = np.transpose(model_dict["trg_word_embedding.word_embedding.weight"])
        else:
            model_dict["decoding_linear.weight"] = model_dict["linear.weight"]

        # To avoid a longer length than training, reset the size of position
        # encoding to max_length
        model_dict["encoder.pos_encoder.weight"] = position_encoding_init(self.max_length, self.d_model)
        model_dict["decoder.pos_encoder.weight"] = position_encoding_init(self.max_length, self.d_model)

        if self.decoding._fuse_qkv:
            for item in self.state_dict():
                if "decoder" in item and "self_attn.q_proj" in item:
                    num_layer = item.split(".")[3]
                    param_type = item.split(".")[-1]

                    model_dict["decoding.slf_q_" + param_type + "_" + num_layer] = np.concatenate(
                        (
                            model_dict[item],
                            model_dict["transformer.decoder.layers." + num_layer + ".self_attn.k_proj." + param_type],
                            model_dict["transformer.decoder.layers." + num_layer + ".self_attn.v_proj." + param_type],
                        ),
                        axis=-1,
                    )

        if self.use_fp16_decoding:
            for item in self.state_dict():
                if "decoder" in item or "decoding.slf" in item:
                    model_dict[item] = np.float16(model_dict[item])
            model_dict["decoding_linear.weight"] = np.float16(model_dict["decoding_linear.weight"])
            model_dict["trg_word_embedding.word_embedding.weight"] = np.float16(
                model_dict["trg_word_embedding.word_embedding.weight"]
            )
            model_dict["trg_pos_embedding.pos_encoder.weight"] = np.float16(
                model_dict["trg_pos_embedding.pos_encoder.weight"]
            )
            model_dict["decoding_linear.bias"] = np.zeros([self.trg_vocab_size], dtype="float16")

        for item in self.state_dict():
            param = self
            attr_list = item.split(".")
            for attr in attr_list:
                param = getattr(param, attr)
            param_name = param.name
            var = paddle.static.global_scope().find_var(param_name).get_tensor()
            var.set(model_dict[item], place)


class TransformerGenerator(paddle.nn.Layer):
    """
    The Transformer model for auto-regressive generation with beam search. It wraps
    `FasterTransformer` and `InferTransformerModel`, and automatically chioces using
    `FasterTransformer` (with jit building) or the slower verison `InferTransformerModel`.

    Args:
        src_vocab_size (int):
            The size of source vocabulary.
        trg_vocab_size (int):
            The size of target vocabulary.
        max_length (int):
            The maximum length of input sequences.
        num_encoder_layers (int):
            The number of sub-layers to be stacked in the encoder.
        num_decoder_layers (int):
            The number of sub-layers to be stacked in the decoder.
        n_head (int):
            The number of head used in multi-head attention.
        d_model (int):
            The dimension for word embeddings, which is also the last dimension of
            the input and output of multi-head attention, position-wise feed-forward
            networks, encoder and decoder.
        d_inner_hid (int):
            Size of the hidden layer in position-wise feed-forward networks.
        dropout (float):
            Dropout rates. Used for pre-process, activation and inside attention.
        weight_sharing (bool):
            Whether to use weight sharing.
        bos_id (int, optional):
            The start token id and also is used as padding id. Defaults to 0.
        eos_id (int, optional):
            The end token id. Defaults to 1.
        beam_size (int, optional):
            The beam width for beam search. Defaults to 4.
        max_out_len (int, optional):
            The maximum output length. Defaults to 256.
        activation (str, optional):
            The activation used in FFN. Defaults to "relu".
        normalize_before (bool, optional):
            Whether to apply pre-normalization. Defaults to True.
        kwargs:
            The key word arguments can be `output_time_major`, `use_ft`, `use_fp16_decoding`,
            `rel_len`, `alpha`:

            - `output_time_major(bool, optional)`: Indicate the data layout of predicted
            Tensor. If `False`, the data layout would be batch major with shape
            `[batch_size, seq_len, beam_size]`. If  `True`, the data layout would
            be time major with shape `[seq_len, batch_size, beam_size]`. Default
            to `False`.

            - `use_ft(bool, optional)`: Whether to use FastGeneration
            for decoding. Default to True if not set.

            - `use_fp16_decoding(bool, optional)`: Whether to use fp16
            for decoding.  Only works when using FastGeneration.

            - `beam_search_version(str, optional)`: Indicating the strategy of
            beam search. It can be 'v1' or 'v2'. 'v2' would select the top
            `beam_size * 2` beams and process the top `beam_size` alive and
            finish beams in them separately, while 'v1' would only select the
            top `beam_size` beams and mix up the alive and finish beams. 'v2' always
            searchs more and get better results, since the alive beams would
            always be `beam_size` while the number of alive beams in `v1` might
            decrease when meeting the end token. However, 'v2' always generates
            longer results thus might do more calculation and be slower.

            - `rel_len(bool, optional)`: Indicating whether `max_out_len` in is
            the length relative to that of source text. Only works in `v2` temporarily.
            It is suggest to set a small `max_out_len` and use `rel_len=True`.
            Default to False if not set.

            - `alpha(float, optional)`: The power number in length penalty
            calculation. Refer to `GNMT <https://arxiv.org/pdf/1609.08144.pdf>`_.
            Only works in `v2` temporarily. Default to 0.6 if not set.

            - diversity_rate(float, optional): Refer to `A Simple, Fast Diverse
            Decoding Algorithm for Neural Generation <https://arxiv.org/abs/1611.08562>`_
            for details. Bigger `diversity_rate` would lead to more diversity.
            if `diversity_rate == 0` is equivalent to naive BeamSearch. Default
            to 0 if not set. **NOTE**: Only works when using FastGeneration
            temporarily.
    """

    def __init__(
        self,
        src_vocab_size,
        trg_vocab_size,
        max_length,
        num_encoder_layers,
        num_decoder_layers,
        n_head,
        d_model,
        d_inner_hid,
        dropout,
        weight_sharing,
        bos_id=0,
        eos_id=1,
        pad_id=None,
        beam_size=4,
        max_out_len=256,
        activation="relu",
        normalize_before=True,
        **kwargs
    ):
        logger.warning("TransformerGenerator is an experimental API and subject to change.")
        # `kwargs` can include output_time_major, use_fp16_decoding, topk, topp.
        # The later three arguments can only work when using FastGeneration,
        # and expose topk, topp later.
        super(TransformerGenerator, self).__init__()
        self.d_model = d_model
        self.max_length = max_length
        self.output_time_major = kwargs.pop("output_time_major", True)
        # Only works for FastGeneration.
        # TODO: original version supports diversity rate.
        diversity_rate = kwargs.pop("diversity_rate", 0.0)
        use_fp16_decoding = kwargs.pop("use_fp16_decoding", False)
        use_ft = kwargs.pop("use_ft", True)
        beam_search_version = kwargs.pop("beam_search_version", "v1")
        rel_len = kwargs.pop("rel_len", False)
        alpha = kwargs.pop("alpha", 0.6)

        # TODO: Faster version needs to update attr to support custom
        # activation and normalize_before which are both aupport in cpp codes.
        if use_ft and activation == "relu" and normalize_before:
            try:
                decoding_strategy = "beam_search_v2" if beam_search_version == "v2" else "beam_search"
                self.transformer = FasterTransformer(
                    src_vocab_size=src_vocab_size,
                    trg_vocab_size=trg_vocab_size,
                    max_length=max_length,
                    num_encoder_layers=num_encoder_layers,
                    num_decoder_layers=num_decoder_layers,
                    n_head=n_head,
                    d_model=d_model,
                    d_inner_hid=d_inner_hid,
                    dropout=dropout,
                    weight_sharing=weight_sharing,
                    bos_id=bos_id,
                    eos_id=eos_id,
                    pad_id=pad_id,
                    beam_size=beam_size,
                    max_out_len=max_out_len,
                    diversity_rate=diversity_rate,
                    decoding_strategy=decoding_strategy,
                    use_fp16_decoding=use_fp16_decoding,
                    rel_len=rel_len,
                    alpha=alpha,
                )
            except Exception:
                logger.warning(
                    "Exception occurs when using FastGeneration. " "The original forward will be involved. "
                )
                if diversity_rate != 0:
                    logger.warning(
                        "diversity_rate would not work since it is only " "supported by FastGeneration temporarily."
                    )
                self.transformer = InferTransformerModel(
                    src_vocab_size=src_vocab_size,
                    trg_vocab_size=trg_vocab_size,
                    max_length=max_length,
                    num_encoder_layers=num_encoder_layers,
                    num_decoder_layers=num_decoder_layers,
                    n_head=n_head,
                    d_model=d_model,
                    d_inner_hid=d_inner_hid,
                    dropout=dropout,
                    weight_sharing=weight_sharing,
                    bos_id=bos_id,
                    eos_id=eos_id,
                    pad_id=pad_id,
                    beam_size=beam_size,
                    max_out_len=max_out_len,
                    output_time_major=self.output_time_major,
                    beam_search_version=beam_search_version,
                    activation=activation,
                    normalize_before=normalize_before,
                    rel_len=rel_len,
                    alpha=alpha,
                )
        else:
            if diversity_rate != 0:
                logger.warning(
                    "diversity_rate would not work since it is only " "supported by FastGeneration temporarily."
                )
            self.transformer = InferTransformerModel(
                src_vocab_size=src_vocab_size,
                trg_vocab_size=trg_vocab_size,
                max_length=max_length,
                num_encoder_layers=num_encoder_layers,
                num_decoder_layers=num_decoder_layers,
                n_head=n_head,
                d_model=d_model,
                d_inner_hid=d_inner_hid,
                dropout=dropout,
                weight_sharing=weight_sharing,
                bos_id=bos_id,
                eos_id=eos_id,
                pad_id=pad_id,
                beam_size=beam_size,
                max_out_len=max_out_len,
                output_time_major=self.output_time_major,
                beam_search_version=beam_search_version,
                activation=activation,
                normalize_before=normalize_before,
                rel_len=rel_len,
                alpha=alpha,
            )

    def forward(self, src_word, trg_word=None):
        r"""
        Performs decoding for transformer model.

        Args:
            src_word (Tensor):
                The ids of source sequence words. It is a tensor with shape
                `[batch_size, source_sequence_length]` and its data type can be
                int or int64.
            trg_word (Tensor):
                The ids of target sequence words. Normally, it should NOT be
                given. If it's given, force decoding with previous output token
                will be trigger. Defaults to None.

        Returns:
            Tensor:
                An int64 tensor shaped indicating the predicted ids. Its shape is
                `[batch_size, seq_len, beam_size]` or `[seq_len, batch_size, beam_size]`
                according to `output_time_major`. While, when using FastGeneration
                and beam search v2, the beam dimension would be doubled to include
                both the top `beam_size` alive and finish beams, thus the tensor
                shape is `[batch_size, seq_len, beam_size * 2]` or `[seq_len, batch_size, beam_size * 2]`.

        Example:
            .. code-block::

                import paddle
                from paddlenlp.ops import TransformerGenerator

                transformer = TransformerGenerator(
                    src_vocab_size=30000,
                    trg_vocab_size=30000,
                    max_length=256,
                    num_encoder_layers=6,
                    num_decoder_layers=6,
                    n_head=8,
                    d_model=512,
                    d_inner_hid=2048,
                    dropout=0.1,
                    weight_sharing=True,
                    bos_id=0,
                    eos_id=1,
                    beam_size=4,
                    max_out_len=256)

                batch_size = 5
                seq_len = 10
                transformer(
                    src_word=paddle.randint(low=3, high=30000, shape=[batch_size, seq_len]))
        """
        out = self.transformer(src_word, trg_word=trg_word)
        # TODO(guosheng): FasterTransformer has an output with layout
        # `[seq_len, batch_size, beam_size]`. While the output layout of
        # original one is `[batch_size, seq_len, beam_size]`. Maybe we need
        # unify them later.
        if not self.output_time_major and isinstance(self.transformer, FasterTransformer):
            out = paddle.transpose(out, [1, 0, 2])
        return out

    def load(self, path=None, state_dict=None):
        if path is None and state_dict is None:
            raise ValueError("Either path or state_dict must be given to load the infer model. ")

        if isinstance(self.transformer, FasterTransformer):
            self.transformer.load(path, state_dict)
        else:
            if state_dict is None:
                state_dict = paddle.load(path)
            self.transformer.load_dict(state_dict)


class FasterOPT(OPTPretrainedModel):
    def __init__(self, model, decoding_lib=None, use_fp16_decoding=False):
        super(FasterOPT, self).__init__()
        self._model = model
        self.use_fp16_decoding = use_fp16_decoding
        self.decoding = InferOptDecoding(model=model, decoding_lib=decoding_lib, use_fp16_decoding=use_fp16_decoding)

    def forward(
        self,
        input_ids,
        seq_len=None,
        attention_mask=None,
        top_k=4,
        top_p=0.0,
        max_length=256,
        bos_token_id=None,
        eos_token_id=None,
        pad_token_id=None,
        forced_eos_token_id=None,
        temperature=0,
        decode_strategy="sample",
        num_return_sequences=1,
        **model_kwargs
    ):
        if input_ids.dtype == paddle.int64:
            input_ids = paddle.cast(input_ids, "int32")

        # change top_p to zero if not using top_p sampling for FT
        if decode_strategy == "greedy_search":
            top_p = 0.0
            top_k = 1
        if top_p == 1.0:
            top_p = 0.0
        if seq_len is None:
            seq_len = paddle.sum(paddle.cast(input_ids != pad_token_id, dtype="int32"), axis=-1, dtype="int32")

            if bos_token_id == pad_token_id and paddle.sum(paddle.any(input_ids == pad_token_id), dtype="int64") > 0:
                seq_len = seq_len + 1

        if num_return_sequences > 1:
            input_ids, model_kwargs = self.expand_inputs_for_generation(
                input_ids, expand_size=num_return_sequences, seq_len=seq_len, attention_mask=attention_mask
            )
            seq_len = model_kwargs["seq_len"]
            attention_mask = model_kwargs.get("attention_mask", None)

        return self.decoding(
            input_ids,
            mem_seq_len=seq_len,
            attention_mask=attention_mask,
            topk=top_k,
            topp=top_p,
            max_out_len=max_length,
            bos_token_id=bos_token_id,
            eos_token_id=eos_token_id,
            pad_token_id=pad_token_id,
            forced_eos_token_id=forced_eos_token_id,
            temperature=temperature,
        )

    def export_params(self, state_to_load, place):
        for item in state_to_load:
            param_data = np.array(state_to_load[item])
            if self.use_fp16_decoding:
                param_data = np.float16(param_data)

            param = self
            attr_list = item.split(".")
            attr_list = ["decoding", "model"] + attr_list
            for attr in attr_list:
                param = getattr(param, attr)
            param_name = param.name
            var = paddle.static.global_scope().find_var(param_name).get_tensor()
            var.set(param_data, place)

    def save_resources(self, tokenizer, path):
        vocab_file = os.path.join(path, "vocab.txt")
        if isinstance(tokenizer, GPTTokenizer):
            with open(vocab_file, "w", encoding="utf-8") as f:
                for token in tokenizer.encoder:
                    f.write(token + "\n")
            merges_file = os.path.join(path, "merges.txt")
            shutil.copyfile(tokenizer._merges_file, merges_file)
        elif isinstance(tokenizer, GPTChineseTokenizer):
            tokenizer.save_resources(path)

    generate = forward


class FasterGPT(GPTPretrainedModel):
    def __init__(self, model, decoding_lib=None, use_fp16_decoding=False):
        super(FasterGPT, self).__init__(model.config)
        self._model = model
        self.use_fp16_decoding = use_fp16_decoding
        self.decoding = InferGptDecoding(model=model, decoding_lib=decoding_lib, use_fp16_decoding=use_fp16_decoding)

    def forward(
        self,
        input_ids,
        seq_len=None,
        attention_mask=None,
        top_k=4,
        top_p=0.0,
        max_length=256,
        bos_token_id=None,
        eos_token_id=None,
        pad_token_id=None,
        forced_eos_token_id=None,
        temperature=0,
        decode_strategy="sample",
        num_return_sequences=1,
        **model_kwargs
    ):
        if input_ids.dtype == paddle.int64:
            input_ids = paddle.cast(input_ids, "int32")

        # change top_p to zero if not using top_p sampling for FT
        if decode_strategy == "greedy_search":
            top_p = 0.0
            top_k = 1
        if top_p == 1.0:
            top_p = 0.0
        if seq_len is None:
            seq_len = paddle.sum(paddle.cast(input_ids != pad_token_id, dtype="int32"), axis=-1, dtype="int32")

            if bos_token_id == pad_token_id and paddle.sum(paddle.any(input_ids == pad_token_id), dtype="int64") > 0:
                seq_len = seq_len + 1

        if num_return_sequences > 1:
            input_ids, model_kwargs = self.expand_inputs_for_generation(
                input_ids, expand_size=num_return_sequences, seq_len=seq_len, attention_mask=attention_mask
            )
            seq_len = model_kwargs["seq_len"]
            attention_mask = model_kwargs.get("attention_mask", None)

        return self.decoding(
            input_ids,
            mem_seq_len=seq_len,
            attention_mask=attention_mask,
            topk=top_k,
            topp=top_p,
            max_out_len=max_length,
            bos_token_id=bos_token_id,
            eos_token_id=eos_token_id,
            pad_token_id=pad_token_id,
            forced_eos_token_id=forced_eos_token_id,
            temperature=temperature,
        )

    def export_params(self, state_to_load, place):
        for item in state_to_load:
            param_data = np.array(state_to_load[item])
            if self.use_fp16_decoding:
                param_data = np.float16(param_data)

            param = self
            attr_list = item.split(".")
            attr_list = ["decoding", "model"] + attr_list
            for attr in attr_list:
                param = getattr(param, attr)
            param_name = param.name
            var = paddle.static.global_scope().find_var(param_name).get_tensor()
            var.set(param_data, place)

    def save_resources(self, tokenizer, path):
        vocab_file = os.path.join(path, "vocab.txt")
        if isinstance(tokenizer, GPTTokenizer):
            with open(vocab_file, "w", encoding="utf-8") as f:
                for token in tokenizer.encoder:
                    f.write(token + "\n")
            merges_file = os.path.join(path, "merges.txt")
            shutil.copyfile(tokenizer._merges_file, merges_file)
        elif isinstance(tokenizer, GPTChineseTokenizer):
            tokenizer.save_resources(path)

    generate = forward


class FasterUnifiedTransformer(UnifiedTransformerPretrainedModel):
    def __init__(self, model, decoding_lib=None, use_fp16_decoding=False):
        super(FasterUnifiedTransformer, self).__init__(model.config)
        self._model = model
        self._use_fp16_decoding = use_fp16_decoding
        self.vocab_size = model.lm_head.decoder_bias.shape[0]
        self.unk_token_id = self._model.unk_token_id
        self.mask_token_id = self._model.mask_token_id
        self.bos_token_id = self._model.bos_token_id
        self.pad_token_id = self._model.pad_token_id
        self.logits_mask = self.generate_logits_mask(use_fp16_decoding)
        self._n_head = self._model.num_attention_heads
        self._hidden_dims = self._model.hidden_size
        self._normalize_before = self._model.normalize_before
        self._size_per_head = self._hidden_dims // self._n_head
        self._n_layer = self._model.num_hidden_layers
        self._hidden_act = self._model.hidden_act

        self.decoding = InferUnifiedDecoding(
            model=self._model,
            decoding_lib=decoding_lib,
            use_fp16_decoding=use_fp16_decoding,
            logits_mask=self.logits_mask,
            n_head=self._n_head,
            hidden_dims=self._hidden_dims,
            size_per_head=self._size_per_head,
            n_layer=self._n_layer,
            unk_id=self.unk_token_id,
            mask_id=self.mask_token_id,
            normalize_before=self._normalize_before,
            hidden_act=self._hidden_act,
        )

    def prepare_inputs_for_generation(
        self, input_ids, token_type_ids, attention_mask, seq_len, position_ids=None, role_ids=None, **kwargs
    ):
        input_ids = input_ids[:, :-1]
        if input_ids.dtype == paddle.int64:
            input_ids = paddle.cast(input_ids, dtype="int32")

        if token_type_ids.dtype == paddle.int64:
            token_type_ids = paddle.cast(token_type_ids, dtype="int32")
        decoder_type_ids = token_type_ids[:, -1:]
        token_type_ids = token_type_ids[:, :-1]

        # TODO(guosheng): attention_mask of UnifiedTransformer uses 0/-INF
        # and is 4D. While now we want to use 1/0 to unify all models and
        # tokenizers.
        attention_mask = attention_mask[:, :, :-1, :-1] if attention_mask.ndim == 4 else attention_mask[:, :-1, :-1]
        attention_mask = paddle.cast(attention_mask == 0, dtype="float16" if self._use_fp16_decoding else "float32")

        seq_len = seq_len - 1
        if seq_len.dtype == paddle.int64:
            seq_len = paddle.cast(seq_len, dtype="int32")

        if position_ids is not None:
            if position_ids.dtype == paddle.int64:
                position_ids = paddle.cast(position_ids, dtype="int32")
            decoder_position_ids = position_ids[:, -1:]
            position_ids = position_ids[:, :-1]
        else:
            decoder_position_ids = None

        field_values = {}
        if role_ids is not None:
            if role_ids.dtype == paddle.int64:
                role_ids = paddle.cast(role_ids, dtype="int32")
            decoder_role_ids = role_ids[:, -1:]
            role_ids = role_ids[:, :-1]
        else:
            decoder_role_ids = None

        field_values["input_ids"] = input_ids
        field_values["token_type_ids"] = token_type_ids
        field_values["attention_mask"] = attention_mask
        field_values["seq_len"] = seq_len
        field_values["decoder_type_ids"] = decoder_type_ids
        field_values["position_ids"] = position_ids
        field_values["decoder_position_ids"] = decoder_position_ids
        field_values["role_ids"] = role_ids
        field_values["decoder_role_ids"] = decoder_role_ids

        return field_values

    def generate_logits_mask(self, use_fp16_decoding):
        # pre-process distribution
        logits_mask = np.zeros(shape=[self.vocab_size], dtype=np.float32)

        if use_fp16_decoding:
            logits_mask[self.unk_token_id] = -1e4
            logits_mask[self.bos_token_id] = -1e4
            logits_mask[self.pad_token_id] = -1e4
        else:
            logits_mask[self.unk_token_id] = -1e9
            logits_mask[self.bos_token_id] = -1e9
            logits_mask[self.pad_token_id] = -1e9

        logits_mask_t = paddle.assign(logits_mask)
        if use_fp16_decoding:
            return paddle.cast(logits_mask_t, dtype="float16")
        else:
            return logits_mask_t

    def forward(
        self,
        input_ids,
        token_type_ids,
        attention_mask,
        seq_len=None,
        role_ids=None,
        position_ids=None,
        max_length=128,
        min_length=0,
        top_k=4,
        top_p=0.0,
        decode_strategy="sampling",
        bos_token_id=None,
        eos_token_id=None,
        pad_token_id=None,
        num_beams=4,
        diversity_rate=0.0,
        temperature=1.0,
        num_return_sequences=1,
        length_penalty=0.6,
        early_stopping=False,
        forced_eos_token_id=None,
        **model_kwargs
    ):

        if seq_len is None:
            assert input_ids is not None, "You have to specify either input_ids when generating seq_len."
            seq_len = paddle.sum(
                paddle.cast(input_ids != self.pad_token_id, dtype="int32"), axis=-1, keepdim=True, dtype="int32"
            )
        if decode_strategy.startswith("beam_search"):
            input_ids, model_kwargs = self.expand_inputs_for_generation(
                input_ids,
                expand_size=num_beams,
                token_type_ids=token_type_ids,
                position_ids=position_ids,
                attention_mask=attention_mask,
                seq_len=seq_len,
                role_ids=role_ids,
            )
        elif decode_strategy == "sampling":
            input_ids, model_kwargs = self.expand_inputs_for_generation(
                input_ids,
                expand_size=num_return_sequences,
                token_type_ids=token_type_ids,
                position_ids=position_ids,
                attention_mask=attention_mask,
                seq_len=seq_len,
                role_ids=role_ids,
            )
        elif decode_strategy == "greedy_search":
            model_kwargs = {
                "token_type_ids": token_type_ids,
                "position_ids": position_ids,
                "attention_mask": attention_mask,
                "seq_len": seq_len,
                "role_ids": role_ids,
            }
        else:
            raise ValueError("Only greedy search, beam search and sampling are supported. ")

        model_inputs = self.prepare_inputs_for_generation(input_ids, **model_kwargs)

        seq_len = model_inputs.pop("seq_len")
        decoder_type_ids = model_inputs.pop("decoder_type_ids")
        role_ids = model_inputs.pop("role_ids", None)
        decoder_role_ids = model_inputs.pop("decoder_role_ids", None)
        position_ids = model_inputs.pop("position_ids", None)
        decoder_position_ids = model_inputs.pop("decoder_position_ids", None)

        return self.decoding(
            input_ids=model_inputs["input_ids"],
            attn_mask=model_inputs["attention_mask"],
            memory_seq_lens=seq_len,
            type_id=model_inputs["token_type_ids"],
            decoder_type_id=decoder_type_ids,
            role_id=role_ids,
            decoder_role_id=decoder_role_ids,
            position_id=position_ids,
            decoder_position_id=decoder_position_ids,
            beam_size=num_beams,
            diversity_rate=diversity_rate,
            topk=top_k,
            topp=top_p,
            decoding_strategy=decode_strategy,
            max_out_len=max_length,
            bos_token_id=bos_token_id,
            eos_token_id=eos_token_id,
            pad_token_id=pad_token_id,
            temperature=temperature,
            length_penalty=length_penalty,
            pos_bias=True,
            forced_eos_token_id=forced_eos_token_id,
            early_stopping=early_stopping,
            min_length=min_length,
        )

    generate = forward


class FasterUNIMOText(UNIMOPretrainedModel):
    def __init__(self, model, decoding_lib=None, use_fp16_decoding=False, **kwargs):
        super(FasterUNIMOText, self).__init__(model.config)
        self._model = model
        self._use_fp16_decoding = use_fp16_decoding
        self.unk_token_id = self._model.unk_token_id
        self.mask_token_id = self._model.mask_token_id
        self.bos_token_id = self._model.bos_token_id
        self.pad_token_id = self._model.pad_token_id
        self.vocab_size = model.lm_head.decoder_bias.shape[0]

        self.logits_mask = self.generate_logits_mask(use_fp16_decoding)
        self._n_head = self._model.num_attention_heads
        self._hidden_dims = self._model.hidden_size
        self._normalize_before = self._model.normalize_before
        self._size_per_head = self._hidden_dims // self._n_head
        self._n_layer = self._model.num_hidden_layers
        self._hidden_act = self._model.hidden_act
        self.trans_out = kwargs.get("trans_out", False)

        self.decoding = InferUnifiedDecoding(
            model=self._model,
            decoding_lib=decoding_lib,
            use_fp16_decoding=use_fp16_decoding,
            logits_mask=self.logits_mask,
            n_head=self._n_head,
            hidden_dims=self._hidden_dims,
            size_per_head=self._size_per_head,
            n_layer=self._n_layer,
            unk_id=self.unk_token_id,
            mask_id=self.mask_token_id,
            normalize_before=self._normalize_before,
            hidden_act=self._hidden_act,
        )

    def prepare_inputs_for_generation(self, input_ids, token_type_ids, attention_mask, **kwargs):
        input_ids = input_ids[:, :-1]
        if input_ids.dtype == paddle.int64:
            input_ids = paddle.cast(input_ids, dtype="int32")

        if token_type_ids.dtype == paddle.int64:
            token_type_ids = paddle.cast(token_type_ids, dtype="int32")
        decoder_type_ids = token_type_ids[:, -1:]
        token_type_ids = token_type_ids[:, :-1]

        attention_mask = attention_mask[:, :, :-1, :-1]
        attention_mask = paddle.cast(attention_mask == 0, dtype="float16" if self._use_fp16_decoding else "float32")

        seq_len = kwargs.get("seq_len") - 1
        if seq_len.dtype == paddle.int64:
            seq_len = paddle.cast(seq_len, dtype="int32")

        return {
            "input_ids": input_ids,
            "token_type_ids": token_type_ids,
            "attention_mask": attention_mask,
            "seq_len": seq_len,
            "decoder_type_ids": decoder_type_ids,
        }

    def generate_logits_mask(self, use_fp16_decoding):
        # pre-process distribution
        logits_mask = np.zeros(shape=[self.vocab_size], dtype=np.float32)

        if use_fp16_decoding:
            logits_mask[self.unk_token_id] = -1e4
            logits_mask[self.bos_token_id] = -1e4
            logits_mask[self.pad_token_id] = -1e4
        else:
            logits_mask[self.unk_token_id] = -1e9
            logits_mask[self.bos_token_id] = -1e9
            logits_mask[self.pad_token_id] = -1e9

        logits_mask_t = paddle.assign(logits_mask)
        if use_fp16_decoding:
            return paddle.cast(logits_mask_t, dtype="float16")
        else:
            return logits_mask_t

    def forward(
        self,
        input_ids,
        token_type_ids,
        attention_mask,
        seq_len=None,
        max_length=128,
        min_length=0,
        top_k=4,
        top_p=0.0,
        num_beams=4,
        decode_strategy="sampling",
        bos_token_id=None,
        eos_token_id=None,
        pad_token_id=None,
        diversity_rate=0.0,
        temperature=1.0,
        num_return_sequences=1,
        length_penalty=0.6,
        early_stopping=False,
        forced_eos_token_id=None,
        position_ids=None,
        **model_kwargs
    ):

        if seq_len is None:
            assert input_ids is not None, "You have to specify either input_ids when generating seq_len."
            seq_len = paddle.sum(paddle.cast(input_ids != self.pad_token_id, dtype="int32"), axis=-1, dtype="int32")
        if decode_strategy.startswith("beam_search"):
            input_ids, model_kwargs = self.expand_inputs_for_generation(
                input_ids,
                expand_size=num_beams,
                token_type_ids=token_type_ids,
                position_ids=position_ids,
                attention_mask=attention_mask,
                seq_len=seq_len,
            )
        elif decode_strategy == "sampling":
            input_ids, model_kwargs = self.expand_inputs_for_generation(
                input_ids,
                expand_size=num_return_sequences,
                token_type_ids=token_type_ids,
                position_ids=position_ids,
                attention_mask=attention_mask,
                seq_len=seq_len,
            )
        elif decode_strategy == "greedy_search":
            model_kwargs = {
                "token_type_ids": token_type_ids,
                "position_ids": position_ids,
                "attention_mask": attention_mask,
                "seq_len": seq_len,
            }
        else:
            raise ValueError("Only greedy search, beam search and sampling are supported. ")

        model_inputs = self.prepare_inputs_for_generation(input_ids, **model_kwargs)
        seq_len = model_inputs.pop("seq_len")
        decoder_type_ids = model_inputs.pop("decoder_type_ids")

        ids, output_scores = self.decoding(
            input_ids=model_inputs["input_ids"],
            attn_mask=model_inputs["attention_mask"],
            memory_seq_lens=seq_len,
            type_id=model_inputs["token_type_ids"],
            decoder_type_id=decoder_type_ids,
            beam_size=num_beams,
            diversity_rate=diversity_rate,
            topk=top_k,
            topp=top_p,
            decoding_strategy=decode_strategy,
            max_out_len=max_length,
            bos_token_id=bos_token_id,
            eos_token_id=eos_token_id,
            pad_token_id=pad_token_id,
            temperature=temperature,
            length_penalty=length_penalty,
            forced_eos_token_id=forced_eos_token_id,
            pos_bias=False,
            early_stopping=early_stopping,
            min_length=min_length,
        )
        if self.trans_out:
            if decode_strategy.startswith("beam_search"):
                ids = ids.transpose([1, 2, 0])
            else:
                ids = ids.transpose([1, 0])
        return ids, output_scores

    generate = forward


class FasterMIRO(UNIMOPretrainedModel):
    def __init__(self, model, decoding_lib=None, use_fp16_decoding=False, **kwargs):
        super(FasterMIRO, self).__init__(model.config)
        self._model = model
        self._use_fp16_decoding = use_fp16_decoding
        self.unk_token_id = self._model.unk_token_id
        self.mask_token_id = self._model.mask_token_id
        self.bos_token_id = self._model.bos_token_id
        self.pad_token_id = self._model.pad_token_id
        self.vocab_size = model.lm_head.decoder_bias.shape[0]

        self.logits_mask = self.generate_logits_mask(use_fp16_decoding)
        self._n_head = self._model.num_attention_heads
        self._hidden_dims = self._model.hidden_size
        self._normalize_before = self._model.normalize_before
        self._size_per_head = self._hidden_dims // self._n_head
        self._n_layer = self._model.num_hidden_layers
        self._hidden_act = self._model.hidden_act
        self.trans_out = kwargs.get("trans_out", False)

        self.decoding = InferMIRODecoding(
            model=self._model,
            decoding_lib=decoding_lib,
            use_fp16_decoding=use_fp16_decoding,
            logits_mask=self.logits_mask,
            n_head=self._n_head,
            hidden_dims=self._hidden_dims,
            size_per_head=self._size_per_head,
            n_layer=self._n_layer,
            unk_id=self.unk_token_id,
            mask_id=self.mask_token_id,
            normalize_before=self._normalize_before,
            hidden_act=self._hidden_act,
        )

    def prepare_inputs_for_generation(self, input_ids, token_type_ids, attention_mask, **kwargs):
        input_ids = input_ids[:, :-1]
        if input_ids.dtype == paddle.int64:
            input_ids = paddle.cast(input_ids, dtype="int32")

        if token_type_ids.dtype == paddle.int64:
            token_type_ids = paddle.cast(token_type_ids, dtype="int32")
        decoder_type_ids = token_type_ids[:, -1:]
        token_type_ids = token_type_ids[:, :-1]

        attention_mask = attention_mask[:, :, :-1, :-1]
        attention_mask = paddle.cast(attention_mask == 0, dtype="float16" if self._use_fp16_decoding else "float32")

        seq_len = kwargs.get("seq_len") - 1
        if seq_len.dtype == paddle.int64:
            seq_len = paddle.cast(seq_len, dtype="int32")

        return {
            "input_ids": input_ids,
            "token_type_ids": token_type_ids,
            "attention_mask": attention_mask,
            "seq_len": seq_len,
            "decoder_type_ids": decoder_type_ids,
        }

    def generate_logits_mask(self, use_fp16_decoding):
        # pre-process distribution
        logits_mask = np.zeros(shape=[self.vocab_size], dtype=np.float32)

        if use_fp16_decoding:
            logits_mask[self.unk_token_id] = -1e4
            logits_mask[self.bos_token_id] = -1e4
            logits_mask[self.pad_token_id] = -1e4
        else:
            logits_mask[self.unk_token_id] = -1e9
            logits_mask[self.bos_token_id] = -1e9
            logits_mask[self.pad_token_id] = -1e9

        logits_mask_t = paddle.assign(logits_mask)
        if use_fp16_decoding:
            return paddle.cast(logits_mask_t, dtype="float16")
        else:
            return logits_mask_t

    def forward(
        self,
        input_ids,
        token_type_ids,
        attention_mask,
        seq_len=None,
        max_length=128,
        min_length=0,
        top_k=4,
        top_p=0.0,
        num_beams=4,
        decode_strategy="sampling",
        bos_token_id=None,
        eos_token_id=None,
        pad_token_id=None,
        diversity_rate=0.0,
        temperature=1.0,
        num_return_sequences=1,
        length_penalty=0.6,
        early_stopping=False,
        forced_eos_token_id=None,
        position_ids=None,
        **model_kwargs
    ):

        if seq_len is None:
            assert input_ids is not None, "You have to specify either input_ids when generating seq_len."
            seq_len = paddle.sum(paddle.cast(input_ids != self.pad_token_id, dtype="int32"), axis=-1, dtype="int32")
        if decode_strategy.startswith("beam_search"):
            input_ids, model_kwargs = self.expand_inputs_for_generation(
                input_ids,
                expand_size=num_beams,
                token_type_ids=token_type_ids,
                position_ids=position_ids,
                attention_mask=attention_mask,
                seq_len=seq_len,
            )
        elif decode_strategy == "sampling":
            input_ids, model_kwargs = self.expand_inputs_for_generation(
                input_ids,
                expand_size=num_return_sequences,
                token_type_ids=token_type_ids,
                position_ids=position_ids,
                attention_mask=attention_mask,
                seq_len=seq_len,
            )
        elif decode_strategy == "greedy_search":
            model_kwargs = {
                "token_type_ids": token_type_ids,
                "position_ids": position_ids,
                "attention_mask": attention_mask,
                "seq_len": seq_len,
            }
        else:
            raise ValueError("Only greedy search, beam search and sampling are supported. ")

        model_inputs = self.prepare_inputs_for_generation(input_ids, **model_kwargs)
        seq_len = model_inputs.pop("seq_len")
        decoder_type_ids = model_inputs.pop("decoder_type_ids")

        ids, output_scores = self.decoding(
            input_ids=model_inputs["input_ids"],
            attn_mask=model_inputs["attention_mask"],
            memory_seq_lens=seq_len,
            type_id=model_inputs["token_type_ids"],
            decoder_type_id=decoder_type_ids,
            beam_size=num_beams,
            diversity_rate=diversity_rate,
            topk=top_k,
            topp=top_p,
            decoding_strategy=decode_strategy,
            max_out_len=max_length,
            bos_token_id=bos_token_id,
            eos_token_id=eos_token_id,
            pad_token_id=pad_token_id,
            temperature=temperature,
            length_penalty=length_penalty,
            forced_eos_token_id=forced_eos_token_id,
            pos_bias=False,
            early_stopping=early_stopping,
            min_length=min_length,
        )
        if self.trans_out:
            if decode_strategy.startswith("beam_search"):
                ids = ids.transpose([1, 2, 0])
            else:
                ids = ids.transpose([1, 0])
        return ids, output_scores

    generate = forward


class FasterBART(BartPretrainedModel):
    enable_faster_encoder_func = enable_fast_encoder

    def __init__(self, model, decoding_lib=None, use_fp16_decoding=False, enable_fast_encoder=True):
        super(FasterBART, self).__init__(model.config)
        self.use_fp16_decoding = use_fp16_decoding
        self._model = model
        if use_fp16_decoding:
            weight_attr = paddle.ParamAttr(initializer=nn.initializer.Assign(model.bart.encoder.embed_tokens.weight))
            model.bart.encoder.embed_tokens = nn.Embedding(
                *model.bart.encoder.embed_tokens.weight.shape, weight_attr=weight_attr
            )
        self.encoder = model.bart.get_encoder()
        self.decoder = model.bart.get_decoder()
        self.pad_token_id = model.bart.config["pad_token_id"]
        self.enable_fast_encoder = enable_fast_encoder

        self.decoding = InferBartDecoding(
            model=self._model, decoding_lib=decoding_lib, use_fp16_decoding=use_fp16_decoding
        )
        if self.enable_fast_encoder:
            # Must use `enable_fast_encoder` in `__init__` when dygraph to static graph.
            self.encoder = FasterBART.enable_faster_encoder_func(self.encoder)

    def get_encoder(self):
        return self.encoder

    def get_decoder(self):
        return self.decoder

    def forward(
        self,
        input_ids=None,
        encoder_output=None,
        seq_len=None,
        num_beams=4,
        top_k=1,
        top_p=0.0,
        decode_strategy="beam_search",
        bos_token_id=None,
        eos_token_id=None,
        pad_token_id=None,
        decoder_start_token_id=None,
        max_length=256,
        diversity_rate=0.0,
        length_penalty=0.6,
        num_return_sequences=1,
        early_stopping=False,
        forced_eos_token_id=None,
        **model_kwargs
    ):

        if encoder_output is None:
            assert input_ids is not None, "You have to specify either input_ids or encoder_output."
            encoder_output = self.prepare_encoder_decoder_kwargs_for_generation(input_ids, model_kwargs)[
                "encoder_output"
            ]
        if seq_len is None:
            assert input_ids is not None, "You have to specify either input_ids when generating seq_len."
            seq_len = paddle.sum(
                paddle.cast(input_ids != self.pad_token_id, dtype="int32"), axis=-1, keepdim=True, dtype="int32"
            )
        if self.use_fp16_decoding:
            encoder_output = paddle.cast(encoder_output, "float16")
        if decode_strategy.startswith("beam_search") and num_beams > 1:
            encoder_output, expanded_kwargs = self.expand_inputs_for_generation(
                encoder_output, expand_size=num_beams, seq_len=seq_len
            )
            seq_len = expanded_kwargs["seq_len"]
        elif decode_strategy == "sampling" and num_return_sequences > 1:
            encoder_output, expanded_kwargs = self.expand_inputs_for_generation(
                encoder_output, expand_size=num_return_sequences, seq_len=seq_len
            )
            seq_len = expanded_kwargs["seq_len"]
        if decoder_start_token_id is not None:
            bos_token_id = decoder_start_token_id
        return self.decoding(
            enc_output=encoder_output,
            memory_seq_lens=seq_len,
            beam_size=num_beams,
            top_k=top_k,
            decoding_strategy=decode_strategy,
            bos_token_id=bos_token_id,
            eos_token_id=eos_token_id,
            pad_token_id=pad_token_id,
            top_p=top_p,
            max_out_len=max_length,
            diversity_rate=diversity_rate,
            alpha=length_penalty,
            early_stopping=early_stopping,
            forced_eos_token_id=forced_eos_token_id,
        )

    generate = forward


class FasterMBART(MBartPretrainedModel):
    enable_faster_encoder_func = enable_fast_encoder

    def __init__(self, model, decoding_lib=None, use_fp16_decoding=False, enable_fast_encoder=False):
        super(FasterMBART, self).__init__(model.config)
        self.use_fp16_decoding = use_fp16_decoding
        self._model = model
        if use_fp16_decoding:
            weight_attr = paddle.ParamAttr(initializer=nn.initializer.Assign(model.mbart.encoder.embed_tokens.weight))
            model.mbart.encoder.embed_tokens = nn.Embedding(
                *model.mbart.encoder.embed_tokens.weight.shape, weight_attr=weight_attr
            )
        self.encoder = model.mbart.get_encoder()
        self.decoder = model.mbart.get_decoder()
        self.pad_token_id = model.mbart.config["pad_token_id"]
        self.enable_fast_encoder = enable_fast_encoder

        self.decoding = InferMBartDecoding(
            model=self._model,
            decoding_lib=decoding_lib,
            use_fp16_decoding=use_fp16_decoding,
            hidden_act=model.mbart.config["activation_function"],
        )

        if self.enable_fast_encoder:
            # Must use `enable_fast_encoder` in `__init__` when dygraph to static graph.
            self.encoder = FasterMBART.enable_faster_encoder_func(self.encoder)

    def get_encoder(self):
        return self.encoder

    def get_decoder(self):
        return self.decoder

    def forward(
        self,
        input_ids=None,
        encoder_output=None,
        seq_len=None,
        forced_bos_token_id=None,
        num_beams=4,
        top_k=1,
        top_p=0.0,
        decode_strategy="beam_search_v3",
        bos_token_id=None,
        eos_token_id=None,
        pad_token_id=None,
        decoder_start_token_id=None,
        max_length=256,
        diversity_rate=0.0,
        length_penalty=0.6,
        temperature=1.0,
        num_return_sequences=1,
        early_stopping=False,
        forced_eos_token_id=None,
        **model_kwargs
    ):

        bos_token_id = bos_token_id if bos_token_id is not None else getattr(self._model, "bos_token_id", None)
        eos_token_id = eos_token_id if eos_token_id is not None else getattr(self._model, "eos_token_id", None)
        pad_token_id = pad_token_id if pad_token_id is not None else getattr(self._model, "pad_token_id", None)
        decoder_start_token_id = (
            decoder_start_token_id
            if decoder_start_token_id is not None
            else getattr(self._model, "decoder_start_token_id", None)
        )

        # (gongenlei) Not enable_fast_encoder temporarily
        if encoder_output is None:
            assert input_ids is not None, "You have to specify either input_ids or encoder_output."
            encoder_output = self.prepare_encoder_decoder_kwargs_for_generation(input_ids, model_kwargs)[
                "encoder_output"
            ]
        batch_size = paddle.shape(encoder_output)[0]
        if seq_len is None:
            assert input_ids is not None, "You have to specify either input_ids when generating seq_len."
            seq_len = paddle.sum(paddle.cast(input_ids != self.pad_token_id, dtype="int32"), axis=-1, dtype="int32")
        if self.use_fp16_decoding:
            encoder_output = paddle.cast(encoder_output, "float16")
        if decode_strategy.startswith("beam_search") and num_beams > 1:
            encoder_output, expanded_kwargs = self.expand_inputs_for_generation(
                encoder_output, expand_size=num_beams, seq_len=seq_len
            )
            seq_len = expanded_kwargs["seq_len"]
        elif decode_strategy == "sampling" and num_return_sequences > 1:
            encoder_output, expanded_kwargs = self.expand_inputs_for_generation(
                encoder_output, expand_size=num_return_sequences, seq_len=seq_len
            )
            seq_len = expanded_kwargs["seq_len"]
        if decoder_start_token_id is not None:
            bos_token_id = decoder_start_token_id

        if not isinstance(forced_bos_token_id, type(input_ids)):
            if forced_bos_token_id is not None:
                if decode_strategy == "sampling":
                    forced_bos_token_id = paddle.full(
                        [batch_size * num_return_sequences, 1], forced_bos_token_id, dtype="int32"
                    )
                else:
                    forced_bos_token_id = paddle.full([batch_size, 1], forced_bos_token_id, dtype="int32")
            else:
                forced_bos_token_id = paddle.zeros([0])
        elif decode_strategy == "sampling":
            num_samples = paddle.shape(encoder_output)[0]
            forced_bos_token_id = paddle.expand(forced_bos_token_id, shape=[num_samples, 1])

        return self.decoding(
            enc_output=encoder_output,
            memory_seq_lens=seq_len,
            beam_size=num_beams,
            trg_word=forced_bos_token_id,
            top_k=top_k,
            top_p=top_p,
            decoding_strategy=decode_strategy,
            diversity_rate=diversity_rate,
            max_out_len=max_length,
            bos_token_id=bos_token_id,
            eos_token_id=eos_token_id,
            pad_token_id=pad_token_id,
            alpha=length_penalty,
            temperature=temperature,
            early_stopping=early_stopping,
        )

    generate = forward


class FasterGPTJ(GPTJPretrainedModel):
    def __init__(self, model, decoding_lib=None, use_fp16_decoding=False):
        super(FasterGPTJ, self).__init__()
        self._model = model
        self.use_fp16_decoding = use_fp16_decoding
        self.decoding = InferGptJDecoding(model=model, decoding_lib=decoding_lib, use_fp16_decoding=use_fp16_decoding)

    def forward(
        self,
        input_ids,
        seq_len=None,
        attention_mask=None,
        top_k=4,
        top_p=0.0,
        min_length=0,
        max_length=256,
        bos_token_id=None,
        eos_token_id=None,
        pad_token_id=None,
        forced_eos_token_id=None,
        temperature=0,
        repetition_penalty=1.0,
        decode_strategy="sampling",
        num_return_sequences=1,
        **model_kwargs
    ):
        if input_ids.dtype == paddle.int64:
            input_ids = paddle.cast(input_ids, "int32")

        # change top_p to zero if not using top_p sampling for FT
        if decode_strategy == "greedy_search":
            top_p = 0.0
            top_k = 1
        if top_p == 1.0:
            top_p = 0.0
        if seq_len is None:
            seq_len = paddle.sum(paddle.cast(input_ids != pad_token_id, dtype="int32"), axis=-1, dtype="int32")

        if num_return_sequences > 1:
            input_ids, model_kwargs = self.expand_inputs_for_generation(
                input_ids, expand_size=num_return_sequences, seq_len=seq_len, attention_mask=attention_mask
            )
            seq_len = model_kwargs["seq_len"]
            attention_mask = model_kwargs.get("attention_mask", None)

        return self.decoding(
            input_ids,
            mem_seq_len=seq_len,
            attention_mask=attention_mask,
            topk=top_k,
            topp=top_p,
            max_out_len=max_length,
            bos_token_id=bos_token_id,
            eos_token_id=eos_token_id,
            pad_token_id=pad_token_id,
            forced_eos_token_id=forced_eos_token_id,
            temperature=temperature,
            repetition_penalty=repetition_penalty,
            min_length=min_length,
        )

    generate = forward


class FasterCodeGen(CodeGenPreTrainedModel):
    def __init__(self, model, decoding_lib=None, use_fp16_decoding=False):
        super(FasterCodeGen, self).__init__(model.config)
        self._model = model
        self.use_fp16_decoding = use_fp16_decoding
        self.decoding = InferGptJDecoding(
            model=model, decoding_lib=decoding_lib, use_fp16_decoding=use_fp16_decoding, transpose_qkv=True
        )

    def forward(
        self,
        input_ids,
        seq_len=None,
        attention_mask=None,
        top_k=4,
        top_p=0.0,
        min_length=0,
        max_length=256,
        bos_token_id=None,
        eos_token_id=None,
        pad_token_id=None,
        forced_eos_token_id=None,
        temperature=0,
        repetition_penalty=1.0,
        decode_strategy="sampling",
        num_return_sequences=1,
        **model_kwargs
    ):
        if input_ids.dtype == paddle.int64:
            input_ids = paddle.cast(input_ids, "int32")

        # change top_p to zero if not using top_p sampling for FT
        if decode_strategy == "greedy_search":
            top_p = 0.0
            top_k = 1
        if top_p == 1.0:
            top_p = 0.0
        if seq_len is None:
            seq_len = paddle.sum(paddle.cast(input_ids != pad_token_id, dtype="int32"), axis=-1, dtype="int32")

        if num_return_sequences > 1:
            input_ids, model_kwargs = self.expand_inputs_for_generation(
                input_ids, expand_size=num_return_sequences, seq_len=seq_len, attention_mask=attention_mask
            )
            seq_len = model_kwargs["seq_len"]
            attention_mask = model_kwargs.get("attention_mask", None)

        return self.decoding(
            input_ids,
            mem_seq_len=seq_len,
            attention_mask=attention_mask,
            topk=top_k,
            topp=top_p,
            max_out_len=max_length,
            bos_token_id=bos_token_id,
            eos_token_id=eos_token_id,
            pad_token_id=pad_token_id,
            forced_eos_token_id=forced_eos_token_id,
            temperature=temperature,
            repetition_penalty=repetition_penalty,
            min_length=min_length,
        )

    generate = forward


class FasterPegasus(PegasusPretrainedModel):
    enable_faster_encoder_func = enable_fast_encoder

    def __init__(self, model, decoding_lib=None, use_fp16_decoding=False, enable_fast_encoder=False, **kwargs):
        super(FasterPegasus, self).__init__()
        self.use_fp16_decoding = use_fp16_decoding
        self._model = model
        self.encoder = model.get_encoder()
        self.decoder = model.get_decoder()
        self.pad_token_id = model.pegasus.config["pad_token_id"]
        self.enable_fast_encoder = enable_fast_encoder
        self.trans_out = kwargs.get("trans_out", False)

        self.decoding = InferPegasusDecoding(
            model=self._model,
            decoding_lib=decoding_lib,
            use_fp16_decoding=use_fp16_decoding,
            hidden_act=model.pegasus.config["activation_function"],
        )

        # TODO(gongenlei): Support faster_encoder
        # if self.enable_fast_encoder:
        #     # Must use `enable_fast_encoder` in `__init__` when dygraph to static graph.
        #     self.encoder = FasterPegasus.enable_faster_encoder_func(self.encoder)

    def get_encoder(self):
        return self.encoder

    def get_decoder(self):
        return self.decoder

    def forward(
        self,
        input_ids=None,
        encoder_output=None,
        seq_len=None,
        min_length=0,
        max_length=256,
        num_beams=4,
        decode_strategy="beam_search_v3",
        decoder_start_token_id=None,
        bos_token_id=None,
        eos_token_id=None,
        pad_token_id=None,
        diversity_rate=0.0,
        length_penalty=0.6,
        top_k=1,
        top_p=0.0,
        temperature=1.0,
        num_return_sequences=1,
        early_stopping=False,
        forced_bos_token_id=None,
        forced_eos_token_id=None,
        **model_kwargs
    ):

        bos_token_id = bos_token_id if bos_token_id is not None else getattr(self._model, "bos_token_id", None)
        eos_token_id = eos_token_id if eos_token_id is not None else getattr(self._model, "eos_token_id", None)
        pad_token_id = pad_token_id if pad_token_id is not None else getattr(self._model, "pad_token_id", None)
        decoder_start_token_id = (
            decoder_start_token_id
            if decoder_start_token_id is not None
            else getattr(self._model, "decoder_start_token_id", None)
        )

        if encoder_output is None:
            assert input_ids is not None, "You have to specify either input_ids or encoder_output."
            encoder_output = self.prepare_encoder_decoder_kwargs_for_generation(input_ids, model_kwargs)[
                "encoder_output"
            ]

        if seq_len is None:
            assert input_ids is not None, "You have to specify either input_ids when generating seq_len."
            seq_len = paddle.sum(paddle.cast(input_ids != self.pad_token_id, dtype="int32"), axis=-1, dtype="int32")
        if self.use_fp16_decoding:
            encoder_output = paddle.cast(encoder_output, "float16")
        if decode_strategy.startswith("beam_search") and num_beams > 1:
            encoder_output, expanded_kwargs = self.expand_inputs_for_generation(
                encoder_output, expand_size=num_beams, seq_len=seq_len
            )
            seq_len = expanded_kwargs["seq_len"]
        elif decode_strategy == "sampling" and num_return_sequences > 1:
            encoder_output, expanded_kwargs = self.expand_inputs_for_generation(
                encoder_output, expand_size=num_return_sequences, seq_len=seq_len
            )
            seq_len = expanded_kwargs["seq_len"]
        if decoder_start_token_id is not None:
            bos_token_id = decoder_start_token_id

        ids = self.decoding(
            enc_output=encoder_output,
            memory_seq_lens=seq_len,
            beam_size=num_beams,
            top_k=top_k,
            top_p=top_p,
            decoding_strategy=decode_strategy,
            max_out_len=max_length,
            min_out_len=min_length,
            diversity_rate=diversity_rate,
            bos_token_id=bos_token_id,
            eos_token_id=eos_token_id,
            pad_token_id=pad_token_id,
            alpha=length_penalty,
            temperature=temperature,
            early_stopping=early_stopping,
            forced_eos_token_id=forced_eos_token_id,
        )

        if self.trans_out:
            if decode_strategy.startswith("beam_search"):
                ids = ids.transpose([1, 2, 0])
            else:
                ids = ids.transpose([1, 0])

        return ids

    generate = forward


class FasterT5(T5PretrainedModel):
    def __init__(self, model, decoding_lib=None, use_fp16_decoding=False):
        super(FasterT5, self).__init__(model.config)
        self.use_fp16_decoding = use_fp16_decoding
        self._model = model
        if use_fp16_decoding:
            weight_attr = paddle.ParamAttr(initializer=nn.initializer.Assign(model.encoder.embed_tokens.weight))
            model.encoder.embed_tokens = nn.Embedding(
                *model.encoder.embed_tokens.weight.shape, weight_attr=weight_attr
            )
        self.encoder = model.t5.get_encoder()
        self.decoder = model.t5.get_decoder()
        self.pad_token_id = model.t5.config["pad_token_id"]

        self.decoding = InferT5Decoding(
            model=self._model, decoding_lib=decoding_lib, use_fp16_decoding=use_fp16_decoding
        )

    def get_encoder(self):
        return self.encoder

    def get_decoder(self):
        return self.decoder

    def forward(
        self,
        input_ids=None,
        encoder_output=None,
        seq_len=None,
        max_length=128,
        min_length=0,
        top_k=4,
        top_p=0.0,
        num_beams=4,
        decode_strategy="sampling",
        decoder_start_token_id=None,
        bos_token_id=None,
        eos_token_id=None,
        pad_token_id=None,
        diversity_rate=0.0,
        temperature=1.0,
        num_return_sequences=1,
        length_penalty=0.6,
        early_stopping=False,
        forced_eos_token_id=None,
        **model_kwargs
    ):

        bos_token_id = bos_token_id if bos_token_id is not None else getattr(self._model, "bos_token_id", None)
        eos_token_id = eos_token_id if eos_token_id is not None else getattr(self._model, "eos_token_id", None)
        pad_token_id = pad_token_id if pad_token_id is not None else getattr(self._model, "pad_token_id", None)

        if encoder_output is None:
            assert input_ids is not None, "You have to specify either input_ids or encoder_output."
            encoder_output = self.prepare_encoder_decoder_kwargs_for_generation(input_ids, model_kwargs)[
                "encoder_output"
            ]

            if isinstance(encoder_output, (list, tuple)):
                encoder_output = encoder_output[0]

        if seq_len is None:
            assert input_ids is not None, "You have to specify either input_ids when generating seq_len."
            seq_len = paddle.sum(paddle.cast(input_ids != self.pad_token_id, dtype="int32"), axis=-1, dtype="int32")
        if self.use_fp16_decoding:
            encoder_output = paddle.cast(encoder_output, "float16")
        if decode_strategy.startswith("beam_search") and num_beams > 1:
            encoder_output, expanded_kwargs = self.expand_inputs_for_generation(
                encoder_output, expand_size=num_beams, seq_len=seq_len
            )
            seq_len = expanded_kwargs["seq_len"]
        elif decode_strategy == "sampling" and num_return_sequences > 1:
            encoder_output, expanded_kwargs = self.expand_inputs_for_generation(
                encoder_output, expand_size=num_return_sequences, seq_len=seq_len
            )
            seq_len = expanded_kwargs["seq_len"]
        if decoder_start_token_id is not None:
            bos_token_id = decoder_start_token_id

        return self.decoding(
            enc_output=encoder_output,
            memory_seq_lens=seq_len,
            beam_size=num_beams,
            top_k=top_k,
            top_p=top_p,
            decoding_strategy=decode_strategy,
            max_out_len=max_length,
            diversity_rate=diversity_rate,
            bos_token_id=bos_token_id,
            eos_token_id=eos_token_id,
            pad_token_id=pad_token_id,
            alpha=length_penalty,
            temperature=temperature,
            early_stopping=early_stopping,
        )

    generate = forward



# coding=utf-8
# Copyright (c) 2022 PaddlePaddle Authors. All Rights Reserved.
# Copyright 2022 The HuggingFace Inc. team. All rights reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Image processor class for DPT."""

import math
from typing import Dict, Iterable, List, Optional, Tuple, Union

import numpy as np
import paddle
import paddle.nn.functional as F
import PIL

from ..image_processing_utils import BaseImageProcessor, BatchFeature, get_size_dict
from ..image_transforms import normalize, rescale, resize, to_channel_dimension_format
from ..image_utils import (
    IMAGENET_STANDARD_MEAN,
    IMAGENET_STANDARD_STD,
    ChannelDimension,
    ImageInput,
    PILImageResampling,
    get_image_size,
    is_batched,
    to_numpy_array,
    valid_images,
)
from ..tokenizer_utils_base import TensorType

__all__ = ["DPTImageProcessor"]


def get_resize_output_image_size(
    input_image: np.ndarray, output_size: Union[int, Iterable[int]], keep_aspect_ratio: bool, multiple: int
) -> Tuple[int, int]:
    def constraint_to_multiple_of(val, multiple, min_val=0, max_val=None):
        x = round(val / multiple) * multiple

        if max_val is not None and x > max_val:
            x = math.floor(val / multiple) * multiple

        if x < min_val:
            x = math.ceil(val / multiple) * multiple

        return x

    output_size = (output_size, output_size) if isinstance(output_size, int) else output_size

    input_height, input_width = get_image_size(input_image)
    output_height, output_width = output_size

    # determine new height and width
    scale_height = output_height / input_height
    scale_width = output_width / input_width

    if keep_aspect_ratio:
        # scale as little as possible
        if abs(1 - scale_width) < abs(1 - scale_height):
            # fit width
            scale_height = scale_width
        else:
            # fit height
            scale_width = scale_height

    new_height = constraint_to_multiple_of(scale_height * input_height, multiple=multiple)
    new_width = constraint_to_multiple_of(scale_width * input_width, multiple=multiple)

    return (new_height, new_width)


class DPTImageProcessor(BaseImageProcessor):
    r"""
    Constructs a DPT image processor.

    Args:
        do_resize (`bool`, *optional*, defaults to `True`):
            Whether to resize the image's (height, width) dimensions. Can be overidden by `do_resize` in `preprocess`.
        size (`Dict[str, int]` *optional*, defaults to `{"height": 384, "width": 384}`):
            Size of the image after resizing. Can be overidden by `size` in `preprocess`.
        keep_aspect_ratio (`bool`, *optional*, defaults to `False`):
            If `True`, the image is resized to the largest possible size such that the aspect ratio is preserved. Can
            be overidden by `keep_aspect_ratio` in `preprocess`.
        ensure_multiple_of (`int`, *optional*, defaults to `1`):
            If `do_resize` is `True`, the image is resized to a size that is a multiple of this value. Can be overidden
            by `ensure_multiple_of` in `preprocess`.
        resample (`PILImageResampling`, *optional*, defaults to `PILImageResampling.BILINEAR`):
            Defines the resampling filter to use if resizing the image. Can be overidden by `resample` in `preprocess`.
        do_rescale (`bool`, *optional*, defaults to `True`):
            Whether to rescale the image by the specified scale `rescale_factor`. Can be overidden by `do_rescale` in
            `preprocess`.
        rescale_factor (`int` or `float`, *optional*, defaults to `1/255`):
            Scale factor to use if rescaling the image. Can be overidden by `rescale_factor` in `preprocess`.
        do_normalize (`bool`, *optional*, defaults to `True`):
            Whether to normalize the image. Can be overridden by the `do_normalize` parameter in the `preprocess`
            method.
        image_mean (`float` or `List[float]`, *optional*, defaults to `IMAGENET_STANDARD_MEAN`):
            Mean to use if normalizing the image. This is a float or list of floats the length of the number of
            channels in the image. Can be overridden by the `image_mean` parameter in the `preprocess` method.
        image_std (`float` or `List[float]`, *optional*, defaults to `IMAGENET_STANDARD_STD`):
            Standard deviation to use if normalizing the image. This is a float or list of floats the length of the
            number of channels in the image. Can be overridden by the `image_std` parameter in the `preprocess` method.
    """

    model_input_names = ["pixel_values"]

    def __init__(
        self,
        do_resize: bool = True,
        size: Dict[str, int] = None,
        resample: PILImageResampling = PILImageResampling.BILINEAR,
        keep_aspect_ratio: bool = False,
        ensure_multiple_of: int = 1,
        do_rescale: bool = True,
        rescale_factor: Union[int, float] = 1 / 255,
        do_normalize: bool = True,
        image_mean: Optional[Union[float, List[float]]] = None,
        image_std: Optional[Union[float, List[float]]] = None,
        **kwargs
    ) -> None:
        super().__init__(**kwargs)
        size = size if size is not None else {"height": 384, "width": 384}
        size = get_size_dict(size)
        self.do_resize = do_resize
        self.size = size
        self.keep_aspect_ratio = keep_aspect_ratio
        self.ensure_multiple_of = ensure_multiple_of
        self.resample = resample
        self.do_rescale = do_rescale
        self.rescale_factor = rescale_factor
        self.do_normalize = do_normalize
        self.image_mean = image_mean if image_mean is not None else IMAGENET_STANDARD_MEAN
        self.image_std = image_std if image_std is not None else IMAGENET_STANDARD_STD

    def resize(
        self,
        image: np.ndarray,
        size: Dict[str, int],
        keep_aspect_ratio: bool = False,
        ensure_multiple_of: int = 1,
        resample: PILImageResampling = PILImageResampling.BICUBIC,
        data_format: Optional[Union[str, ChannelDimension]] = None,
        **kwargs
    ) -> np.ndarray:
        """
        Resize an image to target size `(size["height"], size["width"])`. If `keep_aspect_ratio` is `True`, the image
        is resized to the largest possible size such that the aspect ratio is preserved. If `ensure_multiple_of` is
        set, the image is resized to a size that is a multiple of this value.

        Args:
            image (`np.ndarray`):
                Image to resize.
            size (`Dict[str, int]`):
                Target size of the output image.
            keep_aspect_ratio (`bool`, *optional*, defaults to `False`):
                If `True`, the image is resized to the largest possible size such that the aspect ratio is preserved.
            ensure_multiple_of (`int`, *optional*, defaults to `1`):
                The image is resized to a size that is a multiple of this value.
            resample (`PILImageResampling`, *optional*, defaults to `PILImageResampling.BICUBIC`):
                Defines the resampling filter to use if resizing the image. Otherwise, the image is resized to size
                specified in `size`.
            resample (`PILImageResampling`, *optional*, defaults to `PILImageResampling.BICUBIC`):
                Resampling filter to use when resiizing the image.
            data_format (`str` or `ChannelDimension`, *optional*):
                The channel dimension format of the image. If not provided, it will be the same as the input image.
        """
        size = get_size_dict(size)
        if "height" not in size or "width" not in size:
            raise ValueError(f"The size dictionary must contain the keys 'height' and 'width'. Got {size.keys()}")
        output_size = get_resize_output_image_size(
            image,
            output_size=(size["height"], size["width"]),
            keep_aspect_ratio=keep_aspect_ratio,
            multiple=ensure_multiple_of,
        )
        return resize(image, size=output_size, resample=resample, data_format=data_format, **kwargs)

    def rescale(
        self,
        image: np.ndarray,
        scale: Union[int, float],
        data_format: Optional[Union[str, ChannelDimension]] = None,
        **kwargs
    ):
        """
        Rescale an image by a scale factor. image = image * scale.

        Args:
            image (`np.ndarray`):
                Image to rescale.
            scale (`int` or `float`):
                Scale to apply to the image.
            data_format (`str` or `ChannelDimension`, *optional*):
                The channel dimension format of the image. If not provided, it will be the same as the input image.
        """
        return rescale(image, scale=scale, data_format=data_format, **kwargs)

    def normalize(
        self,
        image: np.ndarray,
        mean: Union[float, List[float]],
        std: Union[float, List[float]],
        data_format: Optional[Union[str, ChannelDimension]] = None,
        **kwargs
    ) -> np.ndarray:
        """
        Normalize an image. image = (image - image_mean) / image_std.

        Args:
            image (`np.ndarray`):
                Image to normalize.
            image_mean (`float` or `List[float]`):
                Image mean.
            image_std (`float` or `List[float]`):
                Image standard deviation.
            data_format (`str` or `ChannelDimension`, *optional*):
                The channel dimension format of the image. If not provided, it will be the same as the input image.
        """
        return normalize(image, mean=mean, std=std, data_format=data_format, **kwargs)

    def preprocess(
        self,
        images: ImageInput,
        do_resize: bool = None,
        size: int = None,
        keep_aspect_ratio: bool = None,
        ensure_multiple_of: int = None,
        resample: PILImageResampling = None,
        do_rescale: bool = None,
        rescale_factor: float = None,
        do_normalize: bool = None,
        image_mean: Optional[Union[float, List[float]]] = None,
        image_std: Optional[Union[float, List[float]]] = None,
        return_tensors: Optional[Union[str, TensorType]] = None,
        data_format: ChannelDimension = ChannelDimension.FIRST,
        **kwargs,
    ) -> PIL.Image.Image:
        """
        Preprocess an image or batch of images.

        Args:
            images (`ImageInput`):
                Image to preprocess.
            do_resize (`bool`, *optional*, defaults to `self.do_resize`):
                Whether to resize the image.
            size (`Dict[str, int]`, *optional*, defaults to `self.size`):
                Size of the image after reszing. If `keep_aspect_ratio` is `True`, the image is resized to the largest
                possible size such that the aspect ratio is preserved. If `ensure_multiple_of` is set, the image is
                resized to a size that is a multiple of this value.
            keep_aspect_ratio (`bool`, *optional*, defaults to `self.keep_aspect_ratio`):
                Whether to keep the aspect ratio of the image. If False, the image will be resized to (size, size). If
                True, the image will be resized to keep the aspect ratio and the size will be the maximum possible.
            ensure_multiple_of (`int`, *optional*, defaults to `self.ensure_multiple_of`):
                Ensure that the image size is a multiple of this value.
            resample (`int`, *optional*, defaults to `self.resample`):
                Resampling filter to use if resizing the image. This can be one of the enum `PILImageResampling`, Only
                has an effect if `do_resize` is set to `True`.
            do_rescale (`bool`, *optional*, defaults to `self.do_rescale`):
                Whether to rescale the image values between [0 - 1].
            rescale_factor (`float`, *optional*, defaults to `self.rescale_factor`):
                Rescale factor to rescale the image by if `do_rescale` is set to `True`.
            do_normalize (`bool`, *optional*, defaults to `self.do_normalize`):
                Whether to normalize the image.
            image_mean (`float` or `List[float]`, *optional*, defaults to `self.image_mean`):
                Image mean.
            image_std (`float` or `List[float]`, *optional*, defaults to `self.image_std`):
                Image standard deviation.
            return_tensors (`str` or `TensorType`, *optional*):
                The type of tensors to return. Can be one of:
                    - Unset: Return a list of `np.ndarray`.
                    - `TensorType.PADDLE` or `'pt'`: Return a batch of type `paddle.Tensor`.
                    - `TensorType.NUMPY` or `'np'`: Return a batch of type `np.ndarray`.
            data_format (`ChannelDimension` or `str`, *optional*, defaults to `ChannelDimension.FIRST`):
                The channel dimension format for the output image. Can be one of:
                    - `ChannelDimension.FIRST`: image in (num_channels, height, width) format.
                    - `ChannelDimension.LAST`: image in (height, width, num_channels) format.
        """
        do_resize = do_resize if do_resize is not None else self.do_resize
        size = size if size is not None else self.size
        size = get_size_dict(size)
        keep_aspect_ratio = keep_aspect_ratio if keep_aspect_ratio is not None else self.keep_aspect_ratio
        ensure_multiple_of = ensure_multiple_of if ensure_multiple_of is not None else self.ensure_multiple_of
        resample = resample if resample is not None else self.resample
        do_rescale = do_rescale if do_rescale is not None else self.do_rescale
        rescale_factor = rescale_factor if rescale_factor is not None else self.rescale_factor
        do_normalize = do_normalize if do_normalize is not None else self.do_normalize
        image_mean = image_mean if image_mean is not None else self.image_mean
        image_std = image_std if image_std is not None else self.image_std

        if not is_batched(images):
            images = [images]

        if not valid_images(images):
            raise ValueError("Invalid image type. Must be of type PIL.Image.Image, numpy.ndarray, " "paddle.Tensor.")

        if do_resize and size is None or resample is None:
            raise ValueError("Size and resample must be specified if do_resize is True.")

        if do_rescale and rescale_factor is None:
            raise ValueError("Rescale factor must be specified if do_rescale is True.")

        if do_normalize and (image_mean is None or image_std is None):
            raise ValueError("Image mean and std must be specified if do_normalize is True.")

        # All transformations expect numpy arrays.
        images = [to_numpy_array(image) for image in images]

        if do_resize:
            images = [self.resize(image=image, size=size, resample=resample) for image in images]

        if do_rescale:
            images = [self.rescale(image=image, scale=rescale_factor) for image in images]

        if do_normalize:
            images = [self.normalize(image=image, mean=image_mean, std=image_std) for image in images]

        images = [to_channel_dimension_format(image, data_format) for image in images]

        data = {"pixel_values": images}
        return BatchFeature(data=data, tensor_type=return_tensors)

    def post_process_semantic_segmentation(self, outputs, target_sizes: List[Tuple] = None):
        """
        Converts the output of [`DPTForSemanticSegmentation`] into semantic segmentation maps. Only supports Paddle.

        Args:
            outputs ([`DPTForSemanticSegmentation`]):
                Raw outputs of the model.
            target_sizes (`List[Tuple]` of length `batch_size`, *optional*):
                List of tuples corresponding to the requested final size (height, width) of each prediction. If unset,
                predictions will not be resized.

        Returns:
            semantic_segmentation: `List[paddle.Tensor]` of length `batch_size`, where each item is a semantic
            segmentation map of shape (height, width) corresponding to the target_sizes entry (if `target_sizes` is
            specified). Each entry of each `paddle.Tensor` correspond to a semantic class id.
        """
        # TODO: add support for other frameworks
        logits = outputs.logits

        # Resize logits and compute semantic segmentation maps
        if target_sizes is not None:
            if len(logits) != len(target_sizes):
                raise ValueError(
                    "Make sure that you pass in as many target sizes as the batch dimension of the logits"
                )

            if paddle.is_tensor(target_sizes):
                target_sizes = target_sizes.numpy()

            semantic_segmentation = []

            for idx in range(len(logits)):
                resized_logits = F.interpolate(
                    logits[idx].unsqueeze(axis=0), size=target_sizes[idx], mode="bilinear", align_corners=False
                )
                semantic_map = resized_logits[0].argmax(axis=0)
                semantic_segmentation.append(semantic_map)
        else:
            semantic_segmentation = logits.argmax(axis=1)
            semantic_segmentation = [semantic_segmentation[i] for i in range(semantic_segmentation.shape[0])]

        return semantic_segmentation



# coding:utf-8
# Copyright (c) 2023  PaddlePaddle Authors. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License"
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
import numpy as np

from ...data import Pad, Tuple
from .base_handler import BaseModelHandler


class CustomModelHandler(BaseModelHandler):
    def __init__(self):
        super().__init__()

    @classmethod
    def process(cls, predictor, tokenizer, data, parameters):
        max_seq_len = 128
        batch_size = 1
        if "max_seq_len" not in parameters:
            max_seq_len = parameters["max_seq_len"]
        if "batch_size" not in parameters:
            batch_size = parameters["batch_size"]
        text = None
        if "text" in data:
            text = data["text"]
        if text is None:
            return {}
        if isinstance(text, str):
            text = [text]
        has_pair = False
        if "text_pair" in data and data["text_pair"] is not None:
            text_pair = data["text_pair"]
            if isinstance(text_pair, str):
                text_pair = [text_pair]
            if len(text) != len(text_pair):
                raise ValueError("The length of text and text_pair must be same.")
            has_pair = True

        # Get the result of tokenizer
        examples = []
        for idx, data in enumerate(text):
            if has_pair:
                result = tokenizer(text=text[idx], text_pair=text_pair[idx], max_length=max_seq_len)
            else:
                result = tokenizer(text=text[idx], max_length=max_seq_len)
            examples.append((result["input_ids"], result["token_type_ids"]))

        # Seperates data into some batches.
        batches = [examples[i : i + batch_size] for i in range(0, len(examples), batch_size)]

        def batchify_fn(samples):
            return Tuple(
                Pad(axis=0, pad_val=tokenizer.pad_token_id, dtype="int64"),
                Pad(axis=0, pad_val=tokenizer.pad_token_type_id, dtype="int64"),
            )(samples)

        results = [[]] * predictor._output_num
        for batch in batches:
            input_ids, token_type_ids = batchify_fn(batch)
            if predictor._predictor_type == "paddle_inference":
                predictor._input_handles[0].copy_from_cpu(input_ids)
                predictor._input_handles[1].copy_from_cpu(token_type_ids)
                predictor._predictor.run()
                output = [output_handle.copy_to_cpu() for output_handle in predictor._output_handles]
                for i, out in enumerate(output):
                    results[i].append(out)
            else:
                predictor._predictor.run(None, {"input_ids": input_ids, "token_type_ids": token_type_ids})
                for i, out in enumerate(output):
                    results[i].append(out)

        # Resolve the logits result and get the predict label and confidence
        results_concat = []
        for i in range(0, len(results)):
            results_concat.append(np.concatenate(results[i], axis=0))
        out_dict = {"logits": results_concat[0].tolist(), "data": data}
        for i in range(1, len(results_concat)):
            out_dict[f"logits_{i}"] = results_concat[i].tolist()
        return out_dict


class ERNIEMHandler(BaseModelHandler):
    def __init__(self):
        super().__init__()

    @classmethod
    def process(cls, predictor, tokenizer, data, parameters):
        max_seq_len = 128
        batch_size = 1
        if "max_seq_len" not in parameters:
            max_seq_len = parameters["max_seq_len"]
        if "batch_size" not in parameters:
            batch_size = parameters["batch_size"]
        text = None
        if "text" in data:
            text = data["text"]
        if text is None:
            return {}
        if isinstance(text, str):
            text = [text]
        has_pair = False
        if "text_pair" in data and data["text_pair"] is not None:
            text_pair = data["text_pair"]
            if isinstance(text_pair, str):
                text_pair = [text_pair]
            if len(text) != len(text_pair):
                raise ValueError("The length of text and text_pair must be same.")
            has_pair = True

        # Get the result of tokenizer
        examples = []
        for idx, data in enumerate(text):
            if has_pair:
                result = tokenizer(text=text[idx], text_pair=text_pair[idx], max_length=max_seq_len)
            else:
                result = tokenizer(text=text[idx], max_length=max_seq_len)
            examples.append(result["input_ids"])

        # Seperates data into some batches.
        batches = [examples[i : i + batch_size] for i in range(0, len(examples), batch_size)]

        def batchify_fn(samples):
            return Pad(axis=0, pad_val=tokenizer.pad_token_id, dtype="int64")(samples)

        results = [[]] * predictor._output_num
        for batch in batches:
            input_ids = batchify_fn(batch)
            if predictor._predictor_type == "paddle_inference":
                predictor._input_handles[0].copy_from_cpu(input_ids)
                predictor._predictor.run()
                output = [output_handle.copy_to_cpu() for output_handle in predictor._output_handles]
                for i, out in enumerate(output):
                    results[i].append(out)
            else:
                predictor._predictor.run(None, {"input_ids": input_ids})
                for i, out in enumerate(output):
                    results[i].append(out)

        # Resolve the logits result and get the predict label and confidence
        results_concat = []
        for i in range(0, len(results)):
            results_concat.append(np.concatenate(results[i], axis=0))
        out_dict = {"logits": results_concat[0].tolist(), "data": data}
        for i in range(1, len(results_concat)):
            out_dict[f"logits_{i}"] = results_concat[i].tolist()
        return out_dict



# Copyright (c) 2021 PaddlePaddle Authors. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
import collections
import json
import os

from paddle.dataset.common import md5file
from paddle.utils.download import get_path_from_url
from paddlenlp.utils.env import DATA_HOME
from . import DatasetBuilder

__all__ = ["NLPCC13EVSAM05HIT"]


class NLPCC13EVSAM05HIT(DatasetBuilder):
    """
    NLPCC13_EVSAM05_HIT is the dataset for dependency parsing.
    The format of this dataset is based on the CoNLL-X style:

        '''
        raw name        definition

        ID              Token counter, starting at 1 for each new sentence.
        FORM            Word form or punctuation symbol.
        LEMMA           Lemma or stem (depending on the particular treebank) of word form, or an underscore if not available.
        CPOSTAG         Coarse-grained part-of-speech tag, where the tagset depends on the treebank.
        POSTAG          Fine-grained part-of-speech tag, where the tagset depends on the treebank.
        FEATS           Unordered set of syntactic and/or morphological features (depending on the particular treebank), or an underscore if not available.
        HEAD            Head of the current token, which is either a value of ID, or zero (’0’) if the token links to the virtual root node of the sentence.
        DEPREL          Dependency relation to the HEAD.
        PHEAD           Projective head of current token, which is either a value of ID or zero (’0’), or an underscore if not available.
        PDEPREL         Dependency relation to the PHEAD, or an underscore if not available.
        '''
    """

    URL = "https://bj.bcebos.com/paddlenlp/datasets/nlpcc13_evsam05_hit.tar.gz"
    MD5 = "5988ede79690dc87aa6e4343b5299944"
    META_INFO = collections.namedtuple("META_INFO", ("file", "md5"))
    SPLITS = {
        "train": META_INFO(os.path.join("nlpcc13_evsam05_hit", "train.conll"), "d82e667950a5e22b18baf595b9feb30f"),
        "dev": META_INFO(os.path.join("nlpcc13_evsam05_hit", "dev.conll"), "b71b08dc85e652769bfbda30b1e352a9"),
        "test": META_INFO(os.path.join("nlpcc13_evsam05_hit", "test.conll"), "784fb9d966a286df5370f7eee4013cf0"),
    }

    def _get_data(self, mode, **kwargs):
        """Downloads dataset."""
        default_root = os.path.join(DATA_HOME, self.__class__.__name__)
        filename, data_hash = self.SPLITS[mode]
        fullname = os.path.join(default_root, filename)
        if not os.path.exists(fullname) or (data_hash and not md5file(fullname) == data_hash):
            get_path_from_url(self.URL, default_root, self.MD5)

        return fullname

    def _read(self, filename, split):
        start = 0
        with open(filename, "r", encoding="utf-8") as f:
            lines = []
            for line in f.readlines():
                if not line.startswith(" "):
                    if not line.startswith("#") and (len(line) == 1 or line.split()[0].isdigit()):
                        lines.append(line.strip())
                else:
                    lines.append("")

        for i, line in enumerate(lines):
            if not line:
                values = list(zip(*[j.split("\t") for j in lines[start:i]]))
                if split == "test":
                    ID, FORM, LEMMA, CPOS, POS, FEATS, HEAD, DEPREL = values
                else:
                    ID, FORM, LEMMA, CPOS, POS, FEATS, HEAD, DEPREL, _, _ = values
                if values:
                    yield {
                        "ID": ID,
                        "FORM": FORM,
                        "LEMMA": LEMMA,
                        "CPOS": CPOS,
                        "POS": POS,
                        "FEATS": FEATS,
                        "HEAD": HEAD,
                        "DEPREL": DEPREL,
                    }
                start = i + 1



# coding:utf-8
# Copyright (c) 2021  PaddlePaddle Authors. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License"
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import glob
import json
import math
import os
import copy
import itertools

import numpy as np
import jieba
from .utils import download_file
from .task import Task
from .lexical_analysis import load_vocab, LacTask
from .named_entity_recognition import NERWordTagTask

usage = r"""
           from paddlenlp import Taskflow 

           # Taskflow base模式
           seg = Taskflow("word_segmentation")
           seg("第十四届全运会在西安举办")
           '''
           ['第十四届', '全运会', '在', '西安', '举办']
           '''

           seg(["第十四届全运会在西安举办", "三亚是一个美丽的城市"])
           '''
           [['第十四届', '全运会', '在', '西安', '举办'], ['三亚', '是', '一个', '美丽', '的', '城市']]
           '''

           # 快速模式分词
           seg = Taskflow("word_segmentation", mode="fast")
           seg("第十四届全运会在西安举办")
           '''
           ['第十四届', '全运会', '在', '西安', '举办']
           '''

           # 精确模式分词
           seg = Taskflow("word_segmentation", mode="accurate")
           seg("李伟拿出具有科学性、可操作性的《陕西省高校管理体制改革实施方案》")
           '''
           ['李伟', '拿出', '具有', '科学性', '、', '可操作性', '的', '《', '陕西省高校管理体制改革实施方案', '》']
           '''
         """


class SegJiebaTask(Task):
    """
    Word Segmentation task for the raw text.
    Args:
        task(string): The name of task.
        model(string): The model name in the task.
        user_dict(string): The user-defined dictionary, default to None.
        kwargs (dict, optional): Additional keyword arguments passed along to the specific task.
    """

    def __init__(self, task, model, user_dict=None, **kwargs):
        super().__init__(task=task, model=model, **kwargs)
        self._user_dict = user_dict
        if self._user_dict:
            jieba.load_userdict(user_dict)

    def _construct_input_spec(self):
        """
        Construct the input spec for the convert dygraph model to static model.
        """
        return None

    def _construct_model(self, model):
        """
        Construct the inference model for the predictor.
        """
        return None

    def _construct_tokenizer(self, model):
        """
        Construct the tokenizer for the predictor.
        """
        return None

    def _preprocess(self, inputs):
        inputs = self._check_input_text(inputs)
        return inputs

    def _postprocess(self, inputs):
        results = inputs if len(inputs) > 1 else inputs[0]
        return results

    def _run_model(self, inputs):
        def cut(string):
            return jieba.lcut(string)

        results = list(map(cut, inputs))
        return results


class SegLACTask(LacTask):
    """
    Segement the sentences to the words using LAC mode.
    Args:
        task(string): The name of task.
        model(string): The model name in the task.
        kwargs (dict, optional): Additional keyword arguments passed along to the specific task.
    """

    def __init__(self, task, model, **kwargs):
        super().__init__(task=task, model="lac", **kwargs)

    def _postprocess(self, inputs):
        """
        The model output is the tag ids, this function will convert the model output to raw text.
        """
        batch_out = []
        lengths = inputs["lens"]
        preds = inputs["result"]
        sents = inputs["text"]
        final_results = []
        for sent_index in range(len(lengths)):
            single_result = {}
            tags = [self._id2tag_dict[str(index)] for index in preds[sent_index][: lengths[sent_index]]]
            sent = sents[sent_index]
            if self._custom:
                self._custom.parse_customization(sent, tags)
            sent_out = []
            tags_out = []
            parital_word = ""
            for ind, tag in enumerate(tags):
                if parital_word == "":
                    parital_word = sent[ind]
                    tags_out.append(tag.split("-")[0])
                    continue
                if tag.endswith("-B") or (tag == "O" and tags[ind - 1] != "O"):
                    sent_out.append(parital_word)
                    tags_out.append(tag.split("-")[0])
                    parital_word = sent[ind]
                    continue
                parital_word += sent[ind]

            if len(sent_out) < len(tags_out):
                sent_out.append(parital_word)
            final_results.append(sent_out)
        final_results = self._auto_joiner(final_results, self.input_mapping)
        final_results = final_results if len(final_results) > 1 else final_results[0]
        return final_results


class SegWordTagTask(NERWordTagTask):
    """
    Segement the sentences to the words using WordTag model.
    Args:
        task(string): The name of task.
        model(string): The model name in the task.
        kwargs (dict, optional): Additional keyword arguments passed along to the specific task.

    """

    def __init__(self, model, task, **kwargs):
        super().__init__(model="wordtag", task=task, **kwargs)

    def _simplify_result(self, results):
        simple_results = []
        for result in results:
            simple_result = []
            if "items" in result:
                for item in result["items"]:
                    simple_result.append(item["item"])
            simple_results.append(simple_result)
        simple_results = simple_results[0] if len(simple_results) == 1 else simple_results
        return simple_results



# Copyright (c) 2021 PaddlePaddle Authors. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import contextlib
import paddle

MODEL_PARALLEL_RNG = "model_parallel_rng"


class RNGStatesTracker:
    """
    Tracker the RNG states.
    """

    def __init__(self):
        # Map from name to the rng state.
        self.states_ = {}
        self.seeds_ = set()

    def add(self, name, seed):
        if seed in self.seeds_:
            raise ValueError("seed {} already exists".format(seed))
        self.seeds_.add(seed)
        if name in self.states_:
            raise ValueError("state {} already exists".format(name))
        orig_rng_state = paddle.get_cuda_rng_state()
        paddle.seed(seed)
        self.states_[name] = paddle.get_cuda_rng_state()
        paddle.set_cuda_rng_state(orig_rng_state)

    @contextlib.contextmanager
    def rng_state(self, name=MODEL_PARALLEL_RNG):
        if name not in self.states_:
            raise ValueError("state {} does not exist".format(name))
        orig_cuda_rng_state = paddle.get_cuda_rng_state()
        paddle.set_cuda_rng_state(self.states_[name])
        try:
            yield
        finally:
            self.states_[name] = paddle.get_cuda_rng_state()
            paddle.set_cuda_rng_state(orig_cuda_rng_state)


RNG_STATE_TRACKER = RNGStatesTracker()


def get_rng_state_tracker():
    return RNG_STATE_TRACKER



# Copyright (c) 2022 PaddlePaddle Authors. All Rights Reserved.
# Copyright 2022 Katherine Crowson and The HuggingFace Team. All rights reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from typing import Union
import math
import numpy as np
import paddle
from scipy import integrate

__all__ = [
    "SchedulerMixin",
    "DDIMScheduler",
    "LMSDiscreteScheduler",
    "PNDMScheduler",
]


def betas_for_alpha_bar(num_diffusion_timesteps, max_beta=0.999):
    """
    Create a beta schedule that discretizes the given alpha_t_bar function, which defines the cumulative product of
    (1-beta) over time from t = [0,1].

    :param num_diffusion_timesteps: the number of betas to produce. :param alpha_bar: a lambda that takes an argument t
    from 0 to 1 and
                      produces the cumulative product of (1-beta) up to that part of the diffusion process.
    :param max_beta: the maximum beta to use; use values lower than 1 to
                     prevent singularities.
    """

    def alpha_bar(time_step):
        return math.cos((time_step + 0.008) / 1.008 * math.pi / 2) ** 2

    betas = []
    for i in range(num_diffusion_timesteps):
        t1 = i / num_diffusion_timesteps
        t2 = (i + 1) / num_diffusion_timesteps
        betas.append(min(1 - alpha_bar(t2) / alpha_bar(t1), max_beta))
    return np.array(betas, dtype=np.float32)


class SchedulerMixin:
    def set_format(self, tensor_format="pd"):
        self.tensor_format = tensor_format
        if tensor_format == "pd":
            for key, value in vars(self).items():
                if isinstance(value, np.ndarray):
                    setattr(self, key, paddle.to_tensor(value))

        return self

    def clip(self, tensor, min_value=None, max_value=None):
        tensor_format = getattr(self, "tensor_format", "pd")

        if tensor_format == "np":
            return np.clip(tensor, min_value, max_value)
        elif tensor_format == "pd":
            return paddle.clip(tensor, min_value, max_value)

        raise ValueError(f"`self.tensor_format`: {self.tensor_format} is not valid.")

    def log(self, tensor):
        tensor_format = getattr(self, "tensor_format", "pd")

        if tensor_format == "np":
            return np.log(tensor)
        elif tensor_format == "pd":
            return paddle.log(tensor)

        raise ValueError(f"`self.tensor_format`: {self.tensor_format} is not valid.")

    def match_shape(
        self,
        values: Union[np.ndarray, paddle.Tensor],
        broadcast_array: Union[np.ndarray, paddle.Tensor],
    ):
        """
        Turns a 1-D array into an array or tensor with len(broadcast_array.shape) dims.

        Args:
            values: an array or tensor of values to extract.
            broadcast_array: an array with a larger shape of K dimensions with the batch
                dimension equal to the length of timesteps.
        Returns:
            a tensor of shape [batch_size, 1, ...] where the shape has K dims.
        """

        tensor_format = getattr(self, "tensor_format", "pd")
        values = values.flatten()

        while len(values.shape) < len(broadcast_array.shape):
            values = values[..., None]

        return values

    def norm(self, tensor):
        tensor_format = getattr(self, "tensor_format", "pd")
        if tensor_format == "np":
            return np.linalg.norm(tensor)
        elif tensor_format == "pd":
            return paddle.norm(tensor.reshape([tensor.shape[0], -1]), axis=-1).mean()

        raise ValueError(f"`self.tensor_format`: {self.tensor_format} is not valid.")

    def randn_like(self, tensor):
        tensor_format = getattr(self, "tensor_format", "pd")
        if tensor_format == "np":
            return np.random.randn(np.shape(tensor))
        elif tensor_format == "pd":
            return paddle.randn(tensor.shape)

        raise ValueError(f"`self.tensor_format`: {self.tensor_format} is not valid.")

    def zeros_like(self, tensor):
        tensor_format = getattr(self, "tensor_format", "pd")
        if tensor_format == "np":
            return np.zeros_like(tensor)
        elif tensor_format == "pd":
            return paddle.zeros_like(tensor)

        raise ValueError(f"`self.tensor_format`: {self.tensor_format} is not valid.")


class LMSDiscreteScheduler(SchedulerMixin):
    def __init__(
        self,
        num_train_timesteps=1000,
        beta_start=0.0001,
        beta_end=0.02,
        beta_schedule="linear",
        trained_betas=None,
        timestep_values=None,
        tensor_format="pd",
    ):
        """
        Linear Multistep Scheduler for discrete beta schedules. Based on the original k-diffusion implementation by
        Katherine Crowson:
        https://github.com/crowsonkb/k-diffusion/blob/481677d114f6ea445aa009cf5bd7a9cdee909e47/k_diffusion/sampling.py#L181
        """
        self.num_train_timesteps = num_train_timesteps
        self.beta_start = beta_start
        self.beta_end = beta_end
        self.beta_schedule = beta_schedule
        self.trained_betas = trained_betas
        self.timestep_values = timestep_values
        self.tensor_format = tensor_format

        if beta_schedule == "linear":
            self.betas = np.linspace(beta_start, beta_end, num_train_timesteps, dtype=np.float32)
        elif beta_schedule == "scaled_linear":
            # this schedule is very specific to the latent diffusion model.
            self.betas = (
                np.linspace(
                    beta_start**0.5,
                    beta_end**0.5,
                    num_train_timesteps,
                    dtype=np.float32,
                )
                ** 2
            )
        else:
            raise NotImplementedError(f"{beta_schedule} does is not implemented for {self.__class__}")

        self.alphas = 1.0 - self.betas
        self.alphas_cumprod = np.cumprod(self.alphas, axis=0)

        self.sigmas = ((1 - self.alphas_cumprod) / self.alphas_cumprod) ** 0.5

        # setable values
        self.num_inference_steps = None
        self.timesteps = np.arange(0, num_train_timesteps)[::-1].copy()
        self.derivatives = []
        self.set_format(tensor_format=tensor_format)

    def get_lms_coefficient(self, order, t, current_order):
        """
        Compute a linear multistep coefficient
        """

        def lms_derivative(tau):
            prod = 1.0
            for k in range(order):
                if current_order == k:
                    continue
                prod *= (tau - self.sigmas[t - k]) / (self.sigmas[t - current_order] - self.sigmas[t - k])
            return prod

        integrated_coeff = integrate.quad(lms_derivative, self.sigmas[t], self.sigmas[t + 1], epsrel=1e-4)[0]

        return integrated_coeff

    def set_timesteps(self, num_inference_steps):
        self.num_inference_steps = num_inference_steps
        self.timesteps = np.linspace(self.num_train_timesteps - 1, 0, num_inference_steps, dtype=float)

        low_idx = np.floor(self.timesteps).astype(int)
        high_idx = np.ceil(self.timesteps).astype(int)
        frac = np.mod(self.timesteps, 1.0)
        sigmas = np.array(((1 - self.alphas_cumprod) / self.alphas_cumprod) ** 0.5)
        sigmas = (1 - frac) * sigmas[low_idx] + frac * sigmas[high_idx]
        self.sigmas = np.concatenate([sigmas, [0.0]])

        self.derivatives = []
        self.set_format(tensor_format=self.tensor_format)

    def step(
        self,
        model_output: Union[paddle.Tensor, np.ndarray],
        timestep: int,
        sample: Union[paddle.Tensor, np.ndarray],
        order: int = 4,
    ):
        sigma = self.sigmas[timestep]

        # 1. compute predicted original sample (x_0) from sigma-scaled predicted noise
        pred_original_sample = sample - sigma * model_output

        # 2. Convert to an ODE derivative
        derivative = (sample - pred_original_sample) / sigma
        self.derivatives.append(derivative)
        if len(self.derivatives) > order:
            self.derivatives.pop(0)

        # 3. Compute linear multistep coefficients
        order = min(timestep + 1, order)
        lms_coeffs = [self.get_lms_coefficient(order, timestep, curr_order) for curr_order in range(order)]

        # 4. Compute previous sample based on the derivatives path
        prev_sample = sample + sum(
            coeff * derivative for coeff, derivative in zip(lms_coeffs, reversed(self.derivatives))
        )

        return {"prev_sample": prev_sample}

    def add_noise(self, original_samples, noise, timesteps):
        sigmas = self.match_shape(self.sigmas[timesteps], noise)
        noisy_samples = original_samples + noise * sigmas
        return noisy_samples

    def __len__(self):
        return self.num_train_timesteps


class PNDMScheduler(SchedulerMixin):
    def __init__(
        self,
        num_train_timesteps=1000,
        beta_start=0.0001,
        beta_end=0.02,
        beta_schedule="linear",
        skip_prk_steps=False,
        tensor_format="pd",
    ):
        self.num_train_timesteps = num_train_timesteps
        self.beta_start = beta_start
        self.beta_end = beta_end
        self.beta_schedule = beta_schedule
        self.skip_prk_steps = skip_prk_steps
        self.tensor_format = tensor_format

        if beta_schedule == "linear":
            self.betas = np.linspace(beta_start, beta_end, num_train_timesteps, dtype=np.float32)
        elif beta_schedule == "scaled_linear":
            # this schedule is very specific to the latent diffusion model.
            self.betas = (
                np.linspace(
                    beta_start**0.5,
                    beta_end**0.5,
                    num_train_timesteps,
                    dtype=np.float32,
                )
                ** 2
            )
        elif beta_schedule == "squaredcos_cap_v2":
            # Glide cosine schedule
            self.betas = betas_for_alpha_bar(num_train_timesteps)
        else:
            raise NotImplementedError(f"{beta_schedule} does is not implemented for {self.__class__}")

        self.alphas = 1.0 - self.betas
        self.alphas_cumprod = np.cumprod(self.alphas, axis=0)

        self.one = np.array(1.0)

        # For now we only support F-PNDM, i.e. the runge-kutta method
        # For more information on the algorithm please take a look at the paper: https://arxiv.org/pdf/2202.09778.pdf
        # mainly at formula (9), (12), (13) and the Algorithm 2.
        self.pndm_order = 4

        # running values
        self.cur_model_output = 0
        self.counter = 0
        self.cur_sample = None
        self.ets = []

        # setable values
        self.num_inference_steps = None
        self._timesteps = np.arange(0, num_train_timesteps)[::-1].copy()
        self._offset = 0
        self.prk_timesteps = None
        self.plms_timesteps = None
        self.timesteps = None
        self.set_format(tensor_format=tensor_format)

    def set_timesteps(self, num_inference_steps, offset=0):
        self.num_inference_steps = num_inference_steps
        self._timesteps = list(
            range(
                0,
                self.num_train_timesteps,
                self.num_train_timesteps // num_inference_steps,
            )
        )
        self._offset = offset
        self._timesteps = np.array([t + self._offset for t in self._timesteps])

        if self.skip_prk_steps:
            # for some models like stable diffusion the prk steps can/should be skipped to
            # produce better results. When using PNDM with `self.skip_prk_steps` the implementation
            # is based on crowsonkb's PLMS sampler implementation: https://github.com/CompVis/latent-diffusion/pull/51
            self.prk_timesteps = np.array([])
            self.plms_timesteps = np.concatenate([self._timesteps[:-1], self._timesteps[-2:-1], self._timesteps[-1:]])[
                ::-1
            ].copy()
        else:
            prk_timesteps = np.array(self._timesteps[-self.pndm_order :]).repeat(2) + np.tile(
                np.array([0, self.num_train_timesteps // num_inference_steps // 2]),
                self.pndm_order,
            )
            self.prk_timesteps = (prk_timesteps[:-1].repeat(2)[1:-1])[::-1].copy()
            self.plms_timesteps = self._timesteps[:-3][
                ::-1
            ].copy()  # we copy to avoid having negative strides which are not supported by paddle.to_tensor

        self.timesteps = np.concatenate([self.prk_timesteps, self.plms_timesteps]).astype(np.int64)

        self.ets = []
        self.counter = 0
        self.set_format(tensor_format=self.tensor_format)

    def step(
        self,
        model_output: Union[paddle.Tensor, np.ndarray],
        timestep: int,
        sample: Union[paddle.Tensor, np.ndarray],
    ):
        if self.counter < len(self.prk_timesteps) and not self.skip_prk_steps:
            return self.step_prk(model_output=model_output, timestep=timestep, sample=sample)
        else:
            return self.step_plms(model_output=model_output, timestep=timestep, sample=sample)

    def step_prk(
        self,
        model_output: Union[paddle.Tensor, np.ndarray],
        timestep: int,
        sample: Union[paddle.Tensor, np.ndarray],
    ):
        """
        Step function propagating the sample with the Runge-Kutta method. RK takes 4 forward passes to approximate the
        solution to the differential equation.
        """
        if self.num_inference_steps is None:
            raise ValueError(
                "Number of inference steps is 'None', you need to run 'set_timesteps' after creating the scheduler"
            )
        diff_to_prev = 0 if self.counter % 2 else self.num_train_timesteps // self.num_inference_steps // 2
        prev_timestep = max(timestep - diff_to_prev, self.prk_timesteps[-1])
        timestep = self.prk_timesteps[self.counter // 4 * 4]

        if self.counter % 4 == 0:
            self.cur_model_output += 1 / 6 * model_output
            self.ets.append(model_output)
            self.cur_sample = sample
        elif (self.counter - 1) % 4 == 0:
            self.cur_model_output += 1 / 3 * model_output
        elif (self.counter - 2) % 4 == 0:
            self.cur_model_output += 1 / 3 * model_output
        elif (self.counter - 3) % 4 == 0:
            model_output = self.cur_model_output + 1 / 6 * model_output
            self.cur_model_output = 0

        # cur_sample should not be `None`
        cur_sample = self.cur_sample if self.cur_sample is not None else sample

        prev_sample = self._get_prev_sample(cur_sample, timestep, prev_timestep, model_output)
        self.counter += 1

        return {"prev_sample": prev_sample}

    def step_plms(
        self,
        model_output: Union[paddle.Tensor, np.ndarray],
        timestep: int,
        sample: Union[paddle.Tensor, np.ndarray],
    ):
        """
        Step function propagating the sample with the linear multi-step method. This has one forward pass with multiple
        times to approximate the solution.
        """
        if self.num_inference_steps is None:
            raise ValueError(
                "Number of inference steps is 'None', you need to run 'set_timesteps' after creating the scheduler"
            )
        if not self.skip_prk_steps and len(self.ets) < 3:
            raise ValueError(
                f"{self.__class__} can only be run AFTER scheduler has been run "
                "in 'prk' mode for at least 12 iterations "
                "See: https://github.com/huggingface/diffusers/blob/main/src/diffusers/pipelines/pipeline_pndm.py "
                "for more information."
            )

        prev_timestep = max(timestep - self.num_train_timesteps // self.num_inference_steps, 0)

        if self.counter != 1:
            self.ets.append(model_output)
        else:
            prev_timestep = timestep
            timestep = timestep + self.num_train_timesteps // self.num_inference_steps

        if len(self.ets) == 1 and self.counter == 0:
            model_output = model_output
            self.cur_sample = sample
        elif len(self.ets) == 1 and self.counter == 1:
            model_output = (model_output + self.ets[-1]) / 2
            sample = self.cur_sample
            self.cur_sample = None
        elif len(self.ets) == 2:
            model_output = (3 * self.ets[-1] - self.ets[-2]) / 2
        elif len(self.ets) == 3:
            model_output = (23 * self.ets[-1] - 16 * self.ets[-2] + 5 * self.ets[-3]) / 12
        else:
            model_output = (1 / 24) * (55 * self.ets[-1] - 59 * self.ets[-2] + 37 * self.ets[-3] - 9 * self.ets[-4])

        prev_sample = self._get_prev_sample(sample, timestep, prev_timestep, model_output)
        self.counter += 1

        return {"prev_sample": prev_sample}

    def _get_prev_sample(self, sample, timestep, timestep_prev, model_output):
        # See formula (9) of PNDM paper https://arxiv.org/pdf/2202.09778.pdf
        # this function computes x_(t−δ) using the formula of (9)
        # Note that x_t needs to be added to both sides of the equation

        # Notation (<variable name> -> <name in paper>
        # alpha_prod_t -> α_t
        # alpha_prod_t_prev -> α_(t−δ)
        # beta_prod_t -> (1 - α_t)
        # beta_prod_t_prev -> (1 - α_(t−δ))
        # sample -> x_t
        # model_output -> e_θ(x_t, t)
        # prev_sample -> x_(t−δ)
        alpha_prod_t = self.alphas_cumprod[timestep + 1 - self._offset]
        alpha_prod_t_prev = self.alphas_cumprod[timestep_prev + 1 - self._offset]
        beta_prod_t = 1 - alpha_prod_t
        beta_prod_t_prev = 1 - alpha_prod_t_prev

        # corresponds to (α_(t−δ) - α_t) divided by
        # denominator of x_t in formula (9) and plus 1
        # Note: (α_(t−δ) - α_t) / (sqrt(α_t) * (sqrt(α_(t−δ)) + sqr(α_t))) =
        # sqrt(α_(t−δ)) / sqrt(α_t))
        sample_coeff = (alpha_prod_t_prev / alpha_prod_t) ** (0.5)

        # corresponds to denominator of e_θ(x_t, t) in formula (9)
        model_output_denom_coeff = alpha_prod_t * beta_prod_t_prev ** (0.5) + (
            alpha_prod_t * beta_prod_t * alpha_prod_t_prev
        ) ** (0.5)

        # full formula (9)
        prev_sample = (
            sample_coeff * sample - (alpha_prod_t_prev - alpha_prod_t) * model_output / model_output_denom_coeff
        )

        return prev_sample

    def add_noise(self, original_samples, noise, timesteps):
        sqrt_alpha_prod = self.alphas_cumprod[timesteps] ** 0.5
        sqrt_alpha_prod = self.match_shape(sqrt_alpha_prod, original_samples)
        sqrt_one_minus_alpha_prod = (1 - self.alphas_cumprod[timesteps]) ** 0.5
        sqrt_one_minus_alpha_prod = self.match_shape(sqrt_one_minus_alpha_prod, original_samples)

        noisy_samples = sqrt_alpha_prod * original_samples + sqrt_one_minus_alpha_prod * noise
        return noisy_samples

    def __len__(self):
        return self.num_train_timesteps


class DDIMScheduler(
    SchedulerMixin,
):
    def __init__(
        self,
        num_train_timesteps=1000,
        beta_start=0.0001,
        beta_end=0.02,
        beta_schedule="linear",
        trained_betas=None,
        timestep_values=None,
        clip_sample=True,
        set_alpha_to_one=True,
        tensor_format="pd",
    ):
        self.num_train_timesteps = num_train_timesteps
        self.beta_start = beta_start
        self.beta_end = beta_end
        self.beta_schedule = beta_schedule
        self.trained_betas = trained_betas
        self.timestep_values = timestep_values
        self.clip_sample = clip_sample
        self.set_alpha_to_one = set_alpha_to_one
        self.tensor_format = tensor_format

        if beta_schedule == "linear":
            self.betas = np.linspace(beta_start, beta_end, num_train_timesteps, dtype=np.float32)
        elif beta_schedule == "scaled_linear":
            # this schedule is very specific to the latent diffusion model.
            self.betas = (
                np.linspace(
                    beta_start**0.5,
                    beta_end**0.5,
                    num_train_timesteps,
                    dtype=np.float32,
                )
                ** 2
            )
        elif beta_schedule == "squaredcos_cap_v2":
            # Glide cosine schedule
            self.betas = betas_for_alpha_bar(num_train_timesteps)
        else:
            raise NotImplementedError(f"{beta_schedule} does is not implemented for {self.__class__}")

        self.alphas = 1.0 - self.betas
        self.alphas_cumprod = np.cumprod(self.alphas, axis=0)

        # At every step in ddim, we are looking into the previous alphas_cumprod
        # For the final step, there is no previous alphas_cumprod because we are already at 0
        # `set_alpha_to_one` decides whether we set this paratemer simply to one or
        # whether we use the final alpha of the "non-previous" one.
        self.final_alpha_cumprod = np.array(1.0) if set_alpha_to_one else self.alphas_cumprod[0]

        # setable values
        self.num_inference_steps = None
        self.timesteps = np.arange(0, num_train_timesteps)[::-1].copy()

        self.tensor_format = tensor_format
        self.set_format(tensor_format=tensor_format)

    def _get_variance(self, timestep, prev_timestep):
        alpha_prod_t = self.alphas_cumprod[timestep]
        alpha_prod_t_prev = self.alphas_cumprod[prev_timestep] if prev_timestep >= 0 else self.final_alpha_cumprod
        beta_prod_t = 1 - alpha_prod_t
        beta_prod_t_prev = 1 - alpha_prod_t_prev

        variance = (beta_prod_t_prev / beta_prod_t) * (1 - alpha_prod_t / alpha_prod_t_prev)

        return variance

    def set_timesteps(self, num_inference_steps, offset=0):
        self.num_inference_steps = num_inference_steps
        self.timesteps = np.arange(
            0,
            self.num_train_timesteps,
            self.num_train_timesteps // self.num_inference_steps,
        )[::-1].copy()
        self.timesteps += offset
        self.set_format(tensor_format=self.tensor_format)

    def step(
        self,
        model_output: Union[paddle.Tensor, np.ndarray],
        timestep: int,
        sample: Union[paddle.Tensor, np.ndarray],
        eta: float = 0.0,
        use_clipped_model_output: bool = False,
    ):
        if self.num_inference_steps is None:
            raise ValueError(
                "Number of inference steps is 'None', you need to run 'set_timesteps' after creating the scheduler"
            )
        # See formulas (12) and (16) of DDIM paper https://arxiv.org/pdf/2010.02502.pdf
        # Ideally, read DDIM paper in-detail understanding

        # Notation (<variable name> -> <name in paper>
        # - pred_noise_t -> e_theta(x_t, t)
        # - pred_original_sample -> f_theta(x_t, t) or x_0
        # - std_dev_t -> sigma_t
        # - eta -> η
        # - pred_sample_direction -> "direction pointingc to x_t"
        # - pred_prev_sample -> "x_t-1"

        # 1. get previous step value (=t-1)
        prev_timestep = timestep - self.num_train_timesteps // self.num_inference_steps

        # 2. compute alphas, betas
        alpha_prod_t = self.alphas_cumprod[timestep]
        alpha_prod_t_prev = self.alphas_cumprod[prev_timestep] if prev_timestep >= 0 else self.final_alpha_cumprod
        beta_prod_t = 1 - alpha_prod_t

        # 3. compute predicted original sample from predicted noise also called
        # "predicted x_0" of formula (12) from https://arxiv.org/pdf/2010.02502.pdf
        pred_original_sample = (sample - beta_prod_t ** (0.5) * model_output) / alpha_prod_t ** (0.5)

        # 4. Clip "predicted x_0"
        if self.clip_sample:
            pred_original_sample = self.clip(pred_original_sample, -1, 1)

        # 5. compute variance: "sigma_t(η)" -> see formula (16)
        # σ_t = sqrt((1 − α_t−1)/(1 − α_t)) * sqrt(1 − α_t/α_t−1)
        variance = self._get_variance(timestep, prev_timestep)
        std_dev_t = eta * variance ** (0.5)

        if use_clipped_model_output:
            # the model_output is always re-derived from the clipped x_0 in Glide
            model_output = (sample - alpha_prod_t ** (0.5) * pred_original_sample) / beta_prod_t ** (0.5)

        # 6. compute "direction pointing to x_t" of formula (12) from https://arxiv.org/pdf/2010.02502.pdf
        pred_sample_direction = (1 - alpha_prod_t_prev - std_dev_t**2) ** (0.5) * model_output

        # 7. compute x_t without "random noise" of formula (12) from https://arxiv.org/pdf/2010.02502.pdf
        prev_sample = alpha_prod_t_prev ** (0.5) * pred_original_sample + pred_sample_direction

        if eta > 0:
            noise = paddle.randn(model_output.shape)
            variance = self._get_variance(timestep, prev_timestep) ** (0.5) * eta * noise

            if not paddle.is_tensor(model_output):
                variance = variance.numpy()

            prev_sample = prev_sample + variance

        return {"prev_sample": prev_sample}

    def add_noise(self, original_samples, noise, timesteps):
        sqrt_alpha_prod = self.alphas_cumprod[timesteps] ** 0.5
        sqrt_alpha_prod = self.match_shape(sqrt_alpha_prod, original_samples)
        sqrt_one_minus_alpha_prod = (1 - self.alphas_cumprod[timesteps]) ** 0.5
        sqrt_one_minus_alpha_prod = self.match_shape(sqrt_one_minus_alpha_prod, original_samples)

        noisy_samples = sqrt_alpha_prod * original_samples + sqrt_one_minus_alpha_prod * noise
        return noisy_samples

    def __len__(self):
        return self.num_train_timesteps



# Copyright (c) 2022 PaddlePaddle Authors. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import os
from typing import Any, Callable, Dict, List, Optional, Tuple

import paddle
import paddle.nn as nn
import paddle.nn.functional as F

from ..data import DataCollator
from ..datasets import MapDataset
from ..losses import RDropLoss
from ..trainer import Trainer, TrainerCallback
from ..trainer.trainer_utils import EvalPrediction, get_scheduler
from ..transformers import PretrainedTokenizer, export_model
from ..utils.log import logger
from .prompt_args import PromptTuningArguments
from .prompt_utils import PromptDataCollatorWithPadding
from .template import AutoTemplate
from .verbalizer import SoftVerbalizer


class PromptTrainer(Trainer):
    """
    PromptTrainer is a feature-complete training and eval loop for PaddleNLP
    on prompt-tuning.
    """

    def __init__(
        self,
        model: nn.Layer,
        tokenizer: PretrainedTokenizer,
        criterion: Optional[nn.Layer] = None,
        args: Optional[PromptTuningArguments] = None,
        data_collator: Optional[DataCollator] = None,
        train_dataset: Optional[MapDataset] = None,
        eval_dataset: Optional[MapDataset] = None,
        compute_metrics: Optional[Callable[[EvalPrediction], Dict]] = None,
        callbacks: Optional[List[TrainerCallback]] = None,
        optimizers: Tuple[paddle.optimizer.Optimizer, paddle.optimizer.lr.LRScheduler] = (None, None),
    ):
        if args is None:
            output_dir = "tmp_trainer"
            logger.info(
                "No `TrainingArguments` passed, initialized with " "output_dir={} by default.".format(output_dir)
            )
            args = PromptTuningArguments(output_dir=output_dir)

        if data_collator is None:
            data_collator = PromptDataCollatorWithPadding(tokenizer, padding=True, return_tensors="pd")

        if criterion is None and (args.use_rgl or args.use_rdrop):
            raise Exception("'To use 'use_rgl', 'use_rdrop', 'criterion' must be specified")

        super(PromptTrainer, self).__init__(
            model=model,
            criterion=criterion,
            args=args,
            data_collator=data_collator,
            train_dataset=train_dataset,
            eval_dataset=eval_dataset,
            tokenizer=tokenizer,
            compute_metrics=compute_metrics,
            callbacks=callbacks,
            optimizers=optimizers,
        )

        self.load_state_dict_from_checkpoint(args.resume_from_checkpoint)

        self.train_dataset = self._map_dataset(self.train_dataset)
        self.eval_dataset = self._map_dataset(self.eval_dataset)

        if self.args.use_rdrop:
            self.rdrop_criterion = RDropLoss()

    def _get_model(self):
        model = self.model
        if isinstance(model, paddle.DataParallel):
            model = model._layers
        return model

    @property
    def template(self):
        return self._get_model().template

    @template.setter
    def template(self, template):
        self._get_model().template = template

    @property
    def verbalizer(self):
        return self._get_model().verbalizer

    @verbalizer.setter
    def verbalizer(self, verbalizer):
        self._get_model().verbalizer = verbalizer

    @property
    def pretrained_model(self):
        return self._get_model().plm

    @pretrained_model.setter
    def pretrained_model(self, model):
        setattr(self._get_model(), "plm", model)

    def _map_dataset(self, dataset: MapDataset):
        if dataset is None:
            return None
        if not isinstance(dataset, MapDataset):
            raise ValueError("Expected `MapDataset` but received {}.".format(type(dataset)))

        def encode_with_template(example):
            return self.template(example)

        return dataset.map(encode_with_template)

    def _prepare_input(self, inputs: Dict):
        return inputs

    def _save(self, output_dir: Optional[str] = None, state_dict: Dict[str, Any] = None):
        super(PromptTrainer, self)._save(output_dir, state_dict)
        output_dir = output_dir if output_dir is not None else self.args.output_dir
        if self.template:
            self.template.save(output_dir)
        if self.verbalizer is not None:
            self.verbalizer.save(output_dir)
        if self.args.save_plm:
            plm_output_dir = os.path.join(output_dir, "plm")
            os.makedirs(plm_output_dir, exist_ok=True)
            self.pretrained_model.save_pretrained(plm_output_dir)

    def load_state_dict_from_checkpoint(self, resume_from_checkpoint: os.PathLike = None):
        if resume_from_checkpoint is not None:
            self.template = AutoTemplate.load_from(
                resume_from_checkpoint, self.tokenizer, self.args.max_seq_length, self._get_model()
            )
        super(PromptTrainer, self).load_state_dict_from_checkpoint(resume_from_checkpoint)

    def get_test_dataloader(self, test_dataset):
        test_dataset = self._map_dataset(test_dataset)
        return super(PromptTrainer, self).get_test_dataloader(test_dataset)

    def create_optimizer(self, lr_scheduler=None):
        """
        Setup the optimizer for both model and prompt parameters.
        """
        if self.optimizer is None:
            optim_cls, optim_kwargs = Trainer.get_optimizer_cls_and_kwargs(self.args)

            plm_parameters = []
            if not self.args.freeze_plm:
                plm_parameters.extend([p for p in self._get_model().plm.parameters() if not p.stop_gradient])

            ppt_parameters = []
            if self.template is not None:
                ppt_parameters.extend([x for n, x in self.template.named_parameters() if not x.stop_gradient])
            if self.verbalizer is not None:
                if isinstance(self.verbalizer, SoftVerbalizer):
                    if not self.args.freeze_plm:
                        plm_parameters.extend(
                            [p for n, p in self.verbalizer.non_head_parameters() if not p.stop_gradient]
                        )
                    ppt_parameters.extend([p for n, p in self.verbalizer.head_parameters()])
                else:
                    ppt_parameters.extend([p for n, p in self.verbalizer.parameters()])

            decay_parameters = [
                p.name for n, p in self._get_model().named_parameters() if not any(nd in n for nd in ["bias", "norm"])
            ]

            if len(plm_parameters) > 0:
                ppt_lr = self.args.ppt_learning_rate / self.args.learning_rate
                lr = self.lr_scheduler if lr_scheduler is None else lr_scheduler
                if len(ppt_parameters) > 0:
                    params = [
                        {"params": plm_parameters},
                        {
                            "params": ppt_parameters,
                            "learning_rate": ppt_lr,
                            "weight_decay": self.args.ppt_weight_decay,
                            "beta1": self.args.ppt_adam_beta1,
                            "beta2": self.args.ppt_adam_beta2,
                            "epsilon": self.args.ppt_adam_epsilon,
                        },
                    ]
                else:
                    params = plm_parameters
            else:
                if self.args.max_steps > 0:
                    max_steps = self.args.max_steps
                else:
                    raise ValueError("Please use `max_steps` to set the maximum training steps.")
                warmup = (
                    self.args.warmup_steps if self.args.warmup_steps > 0 else int(self.args.warmup_ratio * max_steps)
                )
                self.lr_scheduler = get_scheduler(
                    self.args.lr_scheduler_type,
                    learning_rate=self.args.ppt_learning_rate,
                    num_warmup_steps=warmup,
                    num_training_steps=max_steps,
                )
                lr = self.lr_scheduler
                params = ppt_parameters

            self.optimizer = optim_cls(
                learning_rate=lr,
                apply_decay_param_fun=lambda x: x in decay_parameters,
                parameters=params,
                weight_decay=self.args.weight_decay,
                grad_clip=nn.ClipGradByGlobalNorm(self.args.max_grad_norm),
                **optim_kwargs,
            )

        return self.optimizer

    def compute_loss(self, model, inputs, return_outputs=False):
        """
        Compute the total loss for every batch.
        """
        if "labels" not in inputs:
            raise ValueError("Fail to compute loss as `labels` not in {}.".format(inputs))
        labels = inputs["labels"]

        input_dict = inputs.copy()

        if self.criterion is not None:
            # pop labels to move loss computation out of the model
            input_dict.pop("labels")
            input_dict["return_hidden_states"] = True
            logits, hidden_states = model(**input_dict)
            loss = self.criterion(logits, labels)

            if self.args.use_rdrop:
                loss = self._compute_rdrop_loss(model, input_dict, labels, logits, loss)

            if self.args.use_rgl:
                loss += self._compute_rgl_loss(hidden_states, labels)
        else:
            loss, logits, _ = model(**input_dict)

        outputs = (loss, logits)

        return (loss, outputs) if return_outputs else loss

    def _compute_rdrop_loss(self, model, input_dict, labels, outputs, loss):
        re_outputs, _ = model(**input_dict)
        ce_loss = (self.criterion(re_outputs, labels) + loss) * 0.5
        kl_loss = self.rdrop_criterion(outputs, re_outputs)
        loss = ce_loss + self.args.alpha_rdrop * kl_loss
        return loss

    def _compute_rgl_loss(self, embeddings, labels, equal_type="raw"):
        """
        Compute the label consistency loss of sentence embeddings per batch.
        Please refer to https://aclanthology.org/2022.findings-naacl.81/
        for more details.
        """

        def _max_equal(x, y):
            return int(paddle.argmax(x, axis=0) == paddle.argmax(y, axis=0))

        def _raw_equal(x, y):
            return int(x == y)

        if equal_type == "raw":
            equals = _raw_equal
        elif equal_type == "max":
            equals = _max_equal
        else:
            raise ValueError("Unsupported equal type {}.".format(equal_type))
        batch_size = embeddings.shape[0]
        loss = 0
        for i in range(batch_size):
            for j in range(batch_size):
                score = F.cosine_similarity(embeddings[i], embeddings[j], axis=0)
                score = score.unsqueeze(0)
                logits = paddle.concat([(1 - score) * 50, (1 + score) * 50], axis=-1)
                label = paddle.to_tensor([equals(labels[i], labels[j])])
                logits = logits.reshape([-1, logits.shape[-1]])
                loss += F.cross_entropy(logits, label.unsqueeze(0))
        loss = loss / (batch_size * (batch_size - 1))
        loss = loss / 100 * self.args.alpha_rgl

        return loss

    def export_model(self, export_path, input_spec=None, export_type="paddle"):
        os.makedirs(export_path, exist_ok=True)
        self.template.save(export_path)
        if self.verbalizer is not None:
            self.verbalizer.save(export_path)
        if input_spec is None:
            input_spec = self.model.get_input_spec()
        export_model(self.model, input_spec, export_path, export_type)
